

# 零、前言

当面临一个业务问题时，机器学习允许你开发强大而有效的数据驱动的解决方案。最近数据量和来源的爆炸增加了基于数据的解决方案的有效性，因此该领域变得越来越有价值。开发一个机器学习解决方案是有具体要求的，有一些软件和工具支持。一个非常好的选择是使用 R，这是一种开源的统计编程语言，得到了广泛的国际社会的支持。R 结构用于统计分析，国际社会开发最前沿的解决方案。由于这些原因，R 允许你只用几行代码就能开发出强大的机器学习解决方案。

有机器学习教程，它们通常需要一些统计学和计算机科学的基础知识。这本书不仅仅是一个教程。它甚至不需要强大的统计学或计算机科学背景。目标不是为您提供所有技术的完整概述，也不是教您如何构建复杂的解决方案。这本书是一条充满实践例子的道路，为你提供专业知识来建立一个新问题的解决方案。目的是展示这种方法背后最重要的概念，让你对机器学习有深刻的理解，并能够识别和使用新的算法。

# 这本书涵盖了什么

[第 1 章](ch01.html "Chapter 1. Transforming Data into Actions")、*将数据转化为行动*，向您展示新技术如何让您用数据驱动的方法解决业务问题。

[第 2 章](ch02.html "Chapter 2. R – A Powerful Tool for Developing Machine Learning Algorithms")，*R——开发机器学习算法的强大工具*，解释了为什么 R 是机器学习的绝佳选择，并涵盖了软件的基础知识。

[第 3 章](ch03.html "Chapter 3. A Simple Machine Learning Analysis")，*一个简单的机器学习分析*，向你展示了一个简单的机器学习解决方案的例子。

[第 4 章](ch04.html "Chapter 4. Step 1 – Data Exploration and Feature Engineering")，*步骤 1——数据探索和特征工程*，向您展示如何在使用机器学习算法之前清理和转换数据。

[第 5 章](ch05.html "Chapter 5. Step 2 – Applying Machine Learning Techniques")，*第 2 步——应用机器学习技术*，向你展示如何应用机器学习算法来解决问题。

[第 6 章](ch06.html "Chapter 6. Step 3 – Validating the Results")，*第 3 步——验证结果*，向您展示如何测量算法的准确性，以便调整其参数。

[第 7 章](ch07.html "Chapter 7. Overview of Machine Learning Techniques")，*机器学习技术概述*，介绍了机器学习算法的主要分支。

[第 8 章](ch08.html "Chapter 8. Machine Learning Examples Applicable to Businesses")、*适用于企业的机器学习示例*，向你展示如何利用机器学习解决一个商业问题。



# 这本书你需要什么

运行代码所需的唯一软件是 R，最好是 3.0.0+。强烈建议您安装 RStudio 桌面 IDE，尽管这不是必需的。



# 这本书是给谁的

这本书是为那些想要学习如何使用 R 来执行一些机器学习的人准备的，以便从他们的数据中获得洞察力，并找到一些现实生活问题的解决方案。也许你已经知道一点机器学习，但从未使用过 R，或者也许你知道一点 R，但对机器学习很陌生。不管是哪种情况，这本书都会让你快速上手。稍微熟悉一些基本的编程概念会有所帮助，但不要求有任何经验。



# 惯例

在这本书里，你会发现许多区分不同种类信息的文本样式。下面是这些风格的一些例子和它们的含义的解释。

文本中的码字、数据库表名、文件夹名、文件名、文件扩展名、路径名、伪 URL、用户输入和 Twitter 句柄如下所示:“加载包含`random forest`算法的`randomForest`包。”

代码块设置如下:

```r
[default]
arrayFeatures <- names(dtBank)
arrayFeatures <- arrayFeatures[arrayFeatures != 'output']
formulaAll <- paste('output', '~')
formulaAll <- paste(formulaAll, arrayFeatures[1])
for(nameFeature in arrayFeatures[-1]){
  formulaAll <- paste(formulaAll, '+', nameFeature)
}
formulaAll <- formula(formulaAll)
```

当我们希望将您的注意力吸引到代码块的特定部分时，相关的行或项目以粗体显示:

```r
n1 + n2
[1] 5
n1 * n2
[1] 6

```

**新术语**和**重要词语**以粗体显示。

### 注

警告或重要提示出现在这样的框中。

### 提示

提示和技巧是这样出现的。



# 读者反馈

我们随时欢迎读者的反馈。让我们知道你对这本书的看法——你喜欢或不喜欢什么。读者的反馈对我们来说很重要，因为它有助于我们开发出真正让你受益匪浅的图书。

要给我们发送总体反馈，只需给`<[feedback@packtpub.com](mailto:feedback@packtpub.com)>`发电子邮件，并在邮件主题中提及书名。

如果有一个你擅长的主题，并且你有兴趣写一本书或者为一本书投稿，请查看我们在[www.packtpub.com/authors](http://www.packtpub.com/authors)的作者指南。



# 客户支持

既然您已经是 Packt book 的骄傲拥有者，我们有许多东西可以帮助您从购买中获得最大收益。

## 下载示例代码

您可以从您在 http://www.packtpub.com[的账户](http://www.packtpub.com)下载您购买的所有 Packt Publishing 书籍的示例代码文件。如果你在其他地方购买了这本书，你可以访问 http://www.packtpub.com/support[网站](http://www.packtpub.com/support)并注册，这样文件就会直接通过电子邮件发送给你。

## 下载这本书的彩色图片

我们还为您提供了一个 PDF 文件，其中包含本书中使用的截屏的彩色图像。彩色图像将帮助您更好地理解输出中的变化。你可以从[https://www . packtpub . com/sites/default/files/downloads/7740 OS _ colored images 下载这个文件。PDF](https://www.packtpub.com/sites/default/files/downloads/7740OS_coloredimages.PDF) 。

## 勘误表

尽管我们已尽一切努力确保内容的准确性，但错误还是会发生。如果您在我们的某本书中发现了错误——可能是文本或代码中的错误——如果您能向我们报告，我们将不胜感激。这样做，你可以让其他读者免受挫折，并帮助我们改进本书的后续版本。如果您发现任何勘误表，请通过访问[http://www.packtpub.com/submit-errata](http://www.packtpub.com/submit-errata)，选择您的图书，点击**勘误表** **提交** **表格**链接，并输入您的勘误表的详细信息。一旦您的勘误表得到验证，您的提交将被接受，该勘误表将被上传到我们的网站或添加到该标题的勘误表部分下的任何现有勘误表列表中。

要查看之前提交的勘误表，请前往[https://www.packtpub.com/books/content/support](https://www.packtpub.com/books/content/support)，在搜索栏中输入书名。所需信息将出现在**勘误表**部分。

## 引文和参考文献

*   [Chapter 4](ch04.html "Chapter 4. Step 1 – Data Exploration and Feature Engineering"), *Step 1 – Data Exploration and Feature Engineering*, [Chapter 5](ch05.html "Chapter 5. Step 2 – Applying Machine Learning Techniques"), *Step 2 – Applying Machine Learning Techniques*, [Chapter 6](ch06.html "Chapter 6. Step 3 – Validating the Results"), *Step 3 – Validating the Results*, and flag dataset:

    Bache 和 m . lich man(2013 年)。UCI 机器学习知识库[http://archive . ics . UCI . edu/ml]。加州欧文:加州大学信息与计算机科学学院。

*   [Chapter 8](ch08.html "Chapter 8. Machine Learning Examples Applicable to Businesses"), *Machine Learning Examples Applicable to Businesses*, and bank dataset:

    [Moro 等人，2014 年] S. Moro，P. Cortez 和 P. Rita。预测银行电话营销成功的数据驱动方法。决策支持系统，爱思唯尔，62:22-31，2014 年 6 月

## 盗版

互联网上版权材料的盗版是所有媒体都存在的问题。在 Packt，我们非常重视版权和许可证的保护。如果您在互联网上发现我们作品的任何形式的非法拷贝，请立即向我们提供地址或网站名称，以便我们采取补救措施。

请通过`<[copyright@packtpub.com](mailto:copyright@packtpub.com)>`联系我们，并提供可疑盗版材料的链接。

我们感谢您帮助保护我们的作者，以及我们为您带来有价值内容的能力。

## 提问

如果您对本书的任何方面有问题，可以通过`<[questions@packtpub.com](mailto:questions@packtpub.com)>`联系我们，我们将尽最大努力解决问题。