

# 零、前言

数据科学和机器学习是当今技术世界的一些热门词汇。从零售店到财富 500 强公司，每个人都在努力让机器学习为他们提供数据驱动的见解，以发展他们的业务。凭借强大的数据操作功能、机器学习包和活跃的开发人员社区，R 使用户能够构建复杂的机器学习系统来解决现实世界的数据问题。

这本书将带你踏上数据驱动的旅程，从 R 和机器学习的基础开始，逐步建立在解决现实世界问题的项目概念上。

# 这本书涵盖了什么

[第 1 章](ch01.html "Chapter 1. Getting Started with R and Machine Learning")、*R 和机器学习入门*，让你熟悉这本书，帮助你重新熟悉 R 及其基础知识。本章也为你提供了机器学习的简短介绍。

[第 2 章](ch02.html "Chapter 2. Let's Help Machines Learn")，*让我们帮助机器学习*，通过解释构成机器学习基础的概念来深入机器学习。您还将看到各种类型的学习算法，以及一些真实世界的例子。

[第 3 章](ch03.html "Chapter 3. Predicting Customer Shopping Trends with Market Basket Analysis")，*用购物篮分析预测客户购物趋势*，从我们的第一个项目开始，电子商务产品推荐、预测和模式分析，使用各种机器学习技术。本章专门讨论市场购物篮分析和关联规则挖掘，以检测客户购物模式和趋势，并使用这些技术进行产品预测和建议。零售公司和电子商务商店(如 Target、Macy's、Flipkart 和 Amazon)广泛使用这些技术进行产品推荐。

[第 4 章](ch04.html "Chapter 4. Building a Product Recommendation System")，*构建产品推荐系统*，涵盖了我们第一个电子商务产品推荐、预测和模式分析项目的第二部分。这一章专门处理分析电子商务产品评论和不同用户的评级，使用算法和技术，如用户协作过滤，设计一个推荐系统，是生产就绪。

[第 5 章](ch05.html "Chapter 5. Credit Risk Detection and Prediction – Descriptive Analytics")，*信用风险检测和预测——描述性分析*，从我们的第二个项目开始，将机器学习应用于一个复杂的金融场景，我们在其中处理信用风险检测和预测。这一章专门介绍主要目标，查看向银行申请贷款的 1，000 人的金融信用数据集。我们将使用机器学习技术来检测潜在信贷风险的人，如果他们从银行贷款，可能无法偿还贷款，并预测未来的情况。本章还将详细介绍我们的数据集、处理数据时面临的主要挑战、数据集的主要功能以及对数据的探索性和描述性分析。它将总结出适合解决这个问题的最佳机器学习技术。

[第 6 章](ch06.html "Chapter 6. Credit Risk Detection and Prediction – Predictive Analytics")、*信用风险检测和预测——预测分析*，从我们上一章停止的描述性分析开始，着眼于预测分析的使用。在这里，我们专门处理使用几个机器学习算法来检测和预测哪些客户将是潜在的信贷风险，如果他们接受贷款，可能不会偿还给银行。这将最终帮助银行做出数据驱动的决定，决定是否批准贷款。我们将讨论几种监督学习算法，并比较它们的性能。这里还将涉及评估各种机器学习算法的效率和准确性的不同指标。

[第 7 章](ch07.html "Chapter 7. Social Media Analysis – Analyzing Twitter Data")、*社交媒体分析——分析 Twitter 数据*，介绍社交媒体分析的世界。我们首先介绍社交媒体的世界以及通过 Twitter 的 API 收集数据的过程。本章将向您介绍从推文中挖掘有用信息的过程，包括用真实世界的例子可视化 Twitter 数据，推文的聚类和主题建模，当前的挑战和复杂性，以及解决这些问题的策略。我们通过例子展示了如何使用 Twitter 数据来计算一些强大的度量。

[第 8 章](ch08.html "Chapter 8. Sentiment Analysis of Twitter Data")，*Twitter 数据的情感分析*，基于 Twitter APIs 的知识，致力于一个分析推文中情感的项目。该项目提出了多种机器学习算法，用于基于推断的情感对推文进行分类。本章还将以比较的方式呈现这些结果，并帮助您理解这些算法的工作原理和结果差异。



# 这本书你需要什么

该软件适用于本书的所有章节:

*   Windows / Mac OS X / Linux
*   R 3.2.0(或更高版本)
*   RStudio 桌面 0.99(或更高版本)

对于硬件，没有具体的要求，因为 R 可以在任何装有 Mac、Linux 或 Windows 的 PC 上运行，但是为了平稳地运行某些迭代算法，最好有至少 4 GB 的物理内存。



# 这本书是给谁的

如果您对使用最先进的技术从数据中挖掘有用的信息以做出数据驱动的决策感兴趣，这是一本适合您的指南。不要求以前有数据科学方面的经验，尽管非常需要 R 的基础知识。机器学习的先验知识会有所帮助，但不是必需的。



# 惯例

在这本书里，你会发现许多区分不同种类信息的文本样式。下面是这些风格的一些例子和它们的含义的解释。

文本中的码字、数据库表名、文件夹名、文件名、文件扩展名、路径名、伪 URL、用户输入和 Twitter 句柄如下所示:“我们可以通过使用`include`指令来包含其他上下文。”

任何命令行输入或输出都按如下方式编写:

```r

# comparing cluster labels with actual iris  species labels.

table(iris$Species, clusters$cluster)

```

**新术语**和**重要词语**以粗体显示。你在屏幕上看到的单词，例如，在菜单或对话框中，出现在文本中的单词是这样的:“从 Twitter 上与关注谁有关的推荐，到网飞上你可能喜欢的其他电影，到 LinkedIn 上你可能感兴趣的工作，推荐引擎无处不在，而不仅仅是在电子商务平台上。”

### 注意

警告或重要提示出现在这样的框中。

### 提示

提示和技巧是这样出现的。



# 读者反馈

我们随时欢迎读者的反馈。让我们知道你对这本书的看法——你喜欢或不喜欢什么。读者的反馈对我们来说很重要，因为它有助于我们开发出真正让你受益匪浅的图书。

要给我们发送总体反馈，只需发送电子邮件`<[feedback@packtpub.com](mailto:feedback@packtpub.com)>`，并在邮件主题中提及书名。

如果有一个你擅长的主题，并且你有兴趣写一本书或者为一本书投稿，请查看我们在[www.packtpub.com/authors](http://www.packtpub.com/authors)的作者指南。



# 客户支持

既然您已经是 Packt book 的骄傲拥有者，我们有许多东西可以帮助您从购买中获得最大收益。

## 下载示例代码

你可以从你在 http://www.packtpub.com[的账户](http://www.packtpub.com)下载本书的示例代码文件。如果你在其他地方购买了这本书，你可以访问 http://www.packtpub.com/support[网站](http://www.packtpub.com/support)并注册，这样文件就会直接通过电子邮件发送给你。

您可以按照以下步骤下载代码文件:

1.  使用您的电子邮件地址和密码登录或注册我们的网站。
2.  将鼠标指针悬停在顶部的**支架**标签上。
3.  点击**代码下载&勘误表**。
4.  在**搜索**框中输入图书名称。
5.  选择您要下载代码文件的书。
6.  从下拉菜单中选择您购买这本书的地方。
7.  点击**代码下载**。

下载文件后，请确保使用最新版本的解压缩或解压文件夹:

*   WinRAR / 7-Zip for Windows
*   适用于 Mac 的 Zipeg / iZip / UnRarX
*   用于 Linux 的 7-Zip / PeaZip

## 下载该 book 的彩色图像

我们还为您提供了一个 PDF 文件，其中包含本书中使用的截图/图表的彩色图像。彩色图像将帮助您更好地理解输出中的变化。您可以从[https://www . packtpub . com/sites/default/files/downloads/rmachinelementbyexample _ color images . pdf](https://www.packtpub.com/sites/default/files/downloads/RMachineLearningByExample_ColorImages.pdf)下载该文件。

## 勘误表

尽管我们已尽一切努力确保内容的准确性，但错误还是会发生。如果您在我们的某本书中发现了错误——可能是文本或代码中的错误——如果您能向我们报告，我们将不胜感激。这样做，你可以让其他读者免受挫折，并帮助我们改进本书的后续版本。如果您发现任何勘误表，请访问[http://www.packtpub.com/submit-errata](http://www.packtpub.com/submit-errata)，选择您的图书，点击**勘误表提交表**链接，并输入您的勘误表的详细信息。一旦您的勘误表得到验证，您的提交将被接受，该勘误表将被上传到我们的网站或添加到该标题的勘误表部分下的任何现有勘误表列表中。

要查看之前提交的勘误表，请前往 https://www.packtpub.com/books/content/support 的[并在搜索栏中输入书名。所需信息将出现在**勘误表**部分。](https://www.packtpub.com/books/content/support)

## 盗版

互联网上版权材料的盗版是所有媒体都存在的问题。在 Packt，我们非常重视版权和许可证的保护。如果您在互联网上发现我们作品的任何形式的非法拷贝，请立即向我们提供地址或网站名称，以便我们采取补救措施。

请通过`<[copyright@packtpub.com](mailto:copyright@packtpub.com)>`联系我们，并提供可疑盗版材料的链接。

我们感谢您帮助保护我们的作者，以及我们为您带来有价值内容的能力。

## 问题

如果您对本书的任何方面有问题，可以通过`<[questions@packtpub.com](mailto:questions@packtpub.com)>`联系我们，我们将尽最大努力解决问题。