

# 一、R 和机器学习入门

这一介绍性的章节将让你从 R 的基础开始，包括各种构造，有用的数据结构，循环和向量化。如果你已经是一个 R 向导，你可以浏览这些部分，然后直接进入下一部分，讨论机器学习作为一个领域实际上代表了什么，以及它包含的主要领域。我们还将讨论每个领域中使用的不同机器学习技术和算法。最后，我们将通过查看 R 中一些最流行的机器学习包来结束，其中一些我们将在后续章节中使用。

如果你是一名数据或机器学习爱好者，你肯定听说过《哈佛商业评论》称数据科学家是 21 世纪 T2 最性感的工作。

### 注意

参考:[https://HBR . org/2012/10/data-scientist-the-sexy-the-job of-the-ST-century/](https://hbr.org/2012/10/data-scientist-the-sexiest-job-of-the-21st-century/)

当前市场对数据科学家有巨大的需求，主要是因为他们的主要工作是从非结构化和结构化数据中收集关键的见解和信息，以帮助他们的业务和组织实现战略性增长。

你们中的一些人可能想知道机器学习或 R 与所有这些有什么关系！嗯，要成为一名成功的数据科学家，你的工具箱中需要的主要工具之一是一种强大的语言，它能够执行复杂的统计计算，处理各种类型的数据并构建模型，这可以帮助你获得以前未知的见解，而 R 是这方面的完美语言！机器学习是成为数据分析师或数据科学家所需技能的基础，这包括使用各种技术来建立模型以从数据中获得洞察力。

这本书将为你提供精通 R 和机器学习所需的一些基本工具，不仅着眼于概念，还将这些概念应用到现实世界的例子中。说够了；现在让我们和 R 一起开始进入机器学习世界的旅程吧！

在本章中，我们将涵盖以下几个方面:

*   钻研 R 的基础知识
*   理解 R 中的数据结构
*   使用函数
*   控制代码流
*   与 R 一起采取进一步措施
*   理解机器学习基础知识
*   让自己熟悉 R 中流行的机器学习包

# 深入研究 R 的基础知识

这里假设你至少熟悉 R 的基础知识或者以前和 R 合作过。因此，我们不会过多地讨论下载和安装。网上有大量的资源提供这方面的大量信息。我推荐你使用 RStudio，它是一个**集成开发环境** ( **IDE** )，比 base R **图形用户界面** ( **GUI** )要好得多。你可以去 https://www.rstudio.com/获得更多的信息。

### 注意

关于 R 项目的细节，你可以访问[https://www.r-project.org/](https://www.r-project.org/)来了解这门语言。除此之外，R 还有一个巨大的奇妙的软件包可供使用，你可以在 https://cran.r-project.org/的[查看与 R 及其软件包相关的一切，那里包含了所有的档案。](https://cran.r-project.org/)

你一定已经熟悉了 R 交互解释器，通常称为**读取-评估-打印** **循环** ( **REPL** )。这个解释器的行为就像任何要求输入的命令行界面一样，以一个`>`字符开始，这表示 R 正在等待您的输入。如果您的输入跨越多行，比如当您正在编写一个函数时，您将在后续的每一行中看到一个`+`提示，这意味着您没有键入完整的表达式，R 要求您提供表达式的其余部分。

R 也可以读取和执行包含命令和函数的完整文件，这些文件保存在扩展名为`.R`的文件中。通常，任何大的应用程序都由几个`.R`文件组成。每个文件在应用程序中都有自己的角色，通常被称为模块。在接下来的几节中，我们将探讨 R 的一些主要特性和功能。

## 使用 R 作为科学计算器

R 中最基本的构造包括变量和算术运算符，它们可用于执行简单的数学运算，如计算器，甚至复杂的统计计算。

```r

> 5 + 6

[1] 11

> 3 * 2

[1] 6

> 1 / 0

[1] Inf

```

记住 R 中的一切都是向量。甚至是前面代码片段中指示的输出结果。它们有一个前导符号[1]，表示它是一个大小为`1`的向量。

你也可以给变量赋值，并像其他编程语言一样操作它们。

```r

> num <- 6

> num ^ 2

[1] 36

> num

[1] 6     # a variable changes value only on re-assignment

> num <- num ^ 2 * 5 + 10 / 3

> num

[1] 183.3333

```

## 对向量进行操作

R 中最基本的数据结构是一个向量。基本上，R 中的任何东西都是一个向量，即使它是一个单一的数字，就像我们在前面的例子中看到的一样！向量基本上是一个序列或一组值。我们可以使用`:`操作符或`c`函数来创建向量，该函数将这些值连接起来创建一个向量。

```r

> x <- 1:5

> x

[1] 1 2 3 4 5

> y <- c(6, 7, 8 ,9, 10)

> y

[1]  6  7  8  9 10

> z <- x + y

> z

[1]  7  9 11 13 15

```

你可以在前面的代码片段中清楚地看到，我们只使用了`+`操作符，没有使用任何循环就把两个向量加在了一起。这就是所谓的矢量化，我们将在后面详细讨论。下面显示了对向量的更多操作:

```r

> c(1,3,5,7,9) * 2

[1]  2  6 10 14 18

> c(1,3,5,7,9) * c(2, 4)

[1]  2 12 10 28 18 # here the second vector gets recycled

```

**输出:**

![Operating on vectors](img/B03459_01_01.jpg)

```r

> factorial(1:5)

[1]   1   2   6  24 120

> exp(2:10)   # exponential function

[1]     7.389056    20.085537    54.598150   148.413159   403.428793  1096.633158

[7]  2980.957987  8103.083928 22026.465795

> cos(c(0, pi/4))   # cosine function

[1] 1.0000000 0.7071068

> sqrt(c(1, 4, 9, 16))

[1] 1 2 3 4

> sum(1:10)

[1] 55

```

你可能会对第二个操作感到困惑，我们试图用一个更大的向量乘一个更小的向量，但我们仍然得到了一个结果！如果你仔细观察，R 也抛出了一个警告。在这种情况下发生的是，由于两个向量大小不相等，在这种情况下较小的向量`c(2, 4)`被回收或重复成为`c(2, 4, 2, 4, 2)`，然后它与第一个向量`c(1, 3, 5, 7 ,9)`相乘，给出最终的结果向量`c(2, 12, 10, 28, 18)`。这里提到的其他函数是 base R 中的标准函数，还有其他几个函数。

### 提示

**下载示例代码**

你可以从你在[http://www.packtpub.com](http://www.packtpub.com)的账户下载本书的示例代码文件。如果你在其他地方购买了这本书，你可以访问 http://www.packtpub.com/support 并注册，文件会直接通过电子邮件发送给你。

您可以按照以下步骤下载代码文件:

*   使用您的电子邮件地址和密码登录或注册我们的网站。
*   将鼠标指针悬停在顶部的**支撑**标签上
*   点击**代码下载&勘误表**
*   在**搜索**框中输入书名
*   选择您要下载代码文件的书
*   从下拉菜单中选择您购买这本书的地方
*   点击**代码下载**

下载文件后，请确保使用最新版本的解压缩或解压文件夹:

*   WinRAR / 7-Zip for Windows
*   适用于 Mac 的 Zipeg / iZip / UnRarX
*   用于 Linux 的 7-Zip / PeaZip

## 特殊值

由于你将在数据分析和机器学习中处理大量杂乱和肮脏的数据，记住 R 中的一些特殊值是很重要的，这样如果其中一个突然出现，你就不会太惊讶了。

```r

> 1 / 0

[1] Inf

> 0 / 0

[1] NaN

> Inf / NaN

[1] NaN

> Inf / Inf

[1] NaN

> log(Inf)

[1] Inf

> Inf + NA

[1] NA

```

这里您应该关注的主要值是代表**无穷大**的`Inf`、`NaN`不是数字的**、**和`NA`表示缺少的值或**不可用的值**。下面的代码片段显示了对这些特殊值的一些逻辑测试及其结果。请记住`TRUE`和`FALSE`是逻辑数据类型值，类似于其他编程语言。

```r

> vec <- c(0, Inf, NaN, NA)

> is.finite(vec)

[1]  TRUE FALSE FALSE FALSE

> is.nan(vec)

[1] FALSE FALSE  TRUE FALSE

> is.na(vec)

[1] FALSE FALSE  TRUE  TRUE

> is.infinite(vec)

[1] FALSE  TRUE FALSE FALSE

```

从它们的名字来看，这些函数是不言自明的。它们清楚地指出哪些值是有限的，哪些是有限的，并分别检查`NaN`和`NA`值。当清理脏数据时，其中一些功能非常有用。



# R 中的数据结构

在这里，我们将看看 R 中最有用的数据结构，并在一些虚构的例子中使用它们，以更好地掌握它们的语法和构造。我们将在这里介绍的主要数据结构包括:

*   向量
*   阵列和矩阵
*   列表
*   数据帧

这些数据结构在 R 内部被广泛使用，并且被各种 R 包和函数使用，包括我们将在后续章节中使用的机器学习函数和算法。因此，了解如何使用这些数据结构来有效地处理数据是至关重要的。

## 载体

正如我们在前面的章节中简单提到的，向量是 r 中最基本的数据类型。我们使用向量来表示任何东西，无论是输入还是输出。我们之前看到了如何创建向量并对其应用数学运算。我们将在这里看到更多的例子。

### 创造载体

在这里，我们将看看初始化向量的方法，其中一些我们之前也做过，使用像`:`这样的运算符和像`c`这样的函数。在下面的代码片段中，我们将使用`seq`函数族以不同的方式初始化向量。

```r

> c(2.5:4.5, 6, 7, c(8, 9, 10), c(12:15))

 [1]  2.5  3.5  4.5  6.0  7.0  8.0  9.0 10.0 12.0 13.0 14.0 15.0

> vector("numeric", 5)

[1] 0 0 0 0 0

> vector("logical", 5)

[1] FALSE FALSE FALSE FALSE FALSE

> logical(5)

[1] FALSE FALSE FALSE FALSE FALSE

> # seq is a function which creates sequences

> seq.int(1,10)

 [1]  1  2  3  4  5  6  7  8  9 10

> seq.int(1,10,2)

[1] 1 3 5 7 9

> seq_len(10)

 [1]  1  2  3  4  5  6  7  8  9 10

```

### 索引和命名向量

我们可以对向量进行的最重要的操作之一是对向量进行子集化和索引，以访问特定的元素，这在我们希望仅对特定的数据点运行一些代码时非常有用。下面的例子展示了一些索引和子集向量的方法:

```r

> vec <- c("R", "Python", "Julia", "Haskell", "Java", "Scala")

> vec[1]

[1] "R"

> vec[2:4]

[1] "Python"  "Julia"   "Haskell"

> vec[c(1, 3, 5)]

[1] "R"     "Julia" "Java" 

> nums <- c(5, 8, 10, NA, 3, 11)

> nums

[1]  5  8 10 NA  3 11

> which.min(nums)   # index of the minimum element

[1] 5

> which.max(nums)   # index of the maximum element

[1] 6

> nums[which.min(nums)]  # the actual minimum element

[1] 3

> nums[which.max(nums)]  # the actual maximum element

[1] 11

```

现在我们来看看如何给向量命名。这基本上是 R 中的一个很棒的特性，在这里你可以标记向量中的每个元素，使它更易读或者更容易解释。有两种方法可以做到这一点，如下例所示:

```r

> c(first=1, second=2, third=3, fourth=4, fifth=5)

```

**输出:**

![Indexing and naming vectors](img/B03459_01_02.jpg)

```r

> positions <- c(1, 2, 3, 4, 5)

> names(positions) 

NULL

> names(positions) <- c("first", "second", "third", "fourth", "fifth")

> positions

```

**输出:**

![Indexing and naming vectors](img/B03459_01_02.jpg)

```r

> names(positions)

[1] "first"  "second" "third"  "fourth" "fifth"

> positions[c("second", "fourth")]

```

**输出:**

![Indexing and naming vectors](img/B03459_01_03.jpg)

因此，你可以看到，有时标注和命名向量变得非常有用，我们也可以使用元素名而不是值来子集化和切片向量。

## 阵列和矩阵

向量是一维的数据结构，这意味着它们只有一个维度，我们可以使用`length`属性得到它们所包含的元素数量。请记住，数组在其他编程语言中也可能有类似的含义，但在 R 中，它们的含义略有不同。基本上，R 中的数组是保存多维数据的数据结构。矩阵只是具有二维的一般数组的特例，即由属性`rows`和`columns`表示。让我们来看看下面的代码片段中的一些例子。

### 创建数组和矩阵

首先我们将创建一个三维的数组。现在很容易在你的屏幕上显示两个维度，但是要更进一步，R 转换数据有特殊的方法。以下示例显示了 R 如何在每个维度中填充数据(列优先),并显示了 4x3x3 数组的最终输出:

```r

> three.dim.array <- array(

+     1:32,    # input data

+     dim = c(4, 3, 3),   # dimensions

+     dimnames = list(    # names of dimensions

+         c("row1", "row2", "row3", "row4"),

+         c("col1", "col2", "col3"),

+         c("first.set", "second.set", "third.set")

+     )

+ )

> three.dim.array

```

**输出:**

![Creating arrays and matrices](img/B03459_01_04.jpg)

就像我前面提到的，矩阵只是数组的一个特例。我们可以使用`matrix`函数创建一个矩阵，如下例所示。请注意，我们使用参数`byrow`在矩阵中按行填充数据，而不是在任何数组或矩阵中使用 R 的默认按列填充。`ncol`和`nrow`参数分别代表列数和行数。

```r

> mat <- matrix(

+     1:24,   # data

+     nrow = 6,  # num of rows

+     ncol = 4,  # num of columns

+     byrow = TRUE,  # fill the elements row-wise

+ )

> mat

```

**输出:**

![Creating arrays and matrices](img/B03459_01_05.jpg)

### 名称和尺寸

就像我们命名的 vectors 和被访问的元素名称一样，将在下面的代码片段中执行类似的操作。您已经在前面的例子中看到了`dimnames`参数的使用。让我们再看一些例子如下:

```r

> dimnames(three.dim.array)

```

**输出:**

![Names and dimensions](img/B03459_01_06.jpg)

```r

> rownames(three.dim.array)

[1] "row1" "row2" "row3" "row4"

> colnames(three.dim.array)

[1] "col1" "col2" "col3"

> dimnames(mat)

NULL

> rownames(mat)

NULL

> rownames(mat) <- c("r1", "r2", "r3", "r4", "r5", "r6")

> colnames(mat) <- c("c1", "c2", "c3", "c4")

> dimnames(mat)

```

**输出:**

![Names and dimensions](img/B03459_01_07.jpg)

```r

> mat

```

**输出:**

![Names and dimensions](img/B03459_01_08.jpg)

要访问与数组和矩阵相关的维度的详细信息，有一些特殊的函数。下面的例子说明了同样的情况:

```r

> dim(three.dim.array)

[1] 4 3 3

> nrow(three.dim.array)

[1] 4

> ncol(three.dim.array)

[1] 3

> length(three.dim.array)  # product of dimensions

[1] 36

> dim(mat)

[1] 6 4

> nrow(mat)

[1] 6

> ncol(mat)

[1] 4

> length(mat)

[1] 24

```

### 矩阵运算

许多机器学习和优化算法将矩阵作为输入数据进行处理。在下一节中，我们将看一些最常见的矩阵运算的例子。

我们首先初始化两个矩阵，然后查看使用函数组合这两个矩阵的方法，例如返回向量的`c`、`rbind`通过`rows,`组合矩阵，以及通过列执行相同操作的`cbind`。

```r

> mat1 <- matrix(

+     1:15,

+     nrow = 5,

+     ncol = 3,

+     byrow = TRUE,

+     dimnames = list(

+         c("M1.r1", "M1.r2", "M1.r3", "M1.r4", "M1.r5")

+         ,c("M1.c1", "M1.c2", "M1.c3")

+     )

+ )

> mat1

```

**输出:**

![Matrix operations](img/B03459_01_09.jpg)

```r

> mat2 <- matrix(

+     16:30,

+     nrow = 5,

+     ncol = 3,

+     byrow = TRUE,

+     dimnames = list(

+         c("M2.r1", "M2.r2", "M2.r3", "M2.r4", "M2.r5"),

+         c("M2.c1", "M2.c2", "M2.c3")

+     )

+ )

> mat2

```

**输出:**

![Matrix operations](img/B03459_01_10.jpg)

```r

> rbind(mat1, mat2)

```

**输出:**

![Matrix operations](img/B03459_01_11.jpg)

```r

> cbind(mat1, mat2)

```

**输出:**

![Matrix operations](img/B03459_01_12.jpg)

```r

> c(mat1, mat2)

```

**输出:**

![Matrix operations](img/B03459_01_13.jpg)

现在我们来看看可以在矩阵上进行的一些重要的算术运算。根据下面的语法，它们中的大多数都是不言自明的:

```r

> mat1 + mat2   # matrix addition

```

**输出:**

![Matrix operations](img/B03459_01_14.jpg)

```r

> mat1 * mat2  # element-wise multiplication

```

**输出:**

![Matrix operations](img/B03459_01_15.jpg)

```r

> tmat2 <- t(mat2)  # transpose

> tmat2

```

**输出:**

![Matrix operations](img/B03459_01_16.jpg)

```r

> mat1 %*% tmat2   # matrix inner product

```

**输出:**

![Matrix operations](img/B03459_01_17.jpg)

```r

> m <- matrix(c(5, -3, 2, 4, 12, -1, 9, 14, 7), nrow = 3, ncol = 3)

> m

```

**输出:**

![Matrix operations](img/B03459_01_18.jpg)

```r

> inv.m <- solve(m)  # matrix inverse

> inv.m

```

**输出:**

![Matrix operations](img/B03459_01_19.jpg)

```r

> round(m %*% inv.m) # matrix * matrix_inverse = identity matrix

```

**输出:**

![Matrix operations](img/B03459_01_20.jpg)

前面的算术运算只是可以应用于矩阵的大量函数和运算符中最受欢迎的一些。这变得很有用，尤其是在线性优化等领域。

## 列表

列表是向量的一种特殊情况，其中向量中的每个元素可以是不同类型的数据结构，甚至是简单的数据类型。如果您以前使用过它，它在某些方面类似于 Python 编程语言中的列表，其中列表指示不同类型的元素，并且每个元素在列表中都有特定的索引。在 R 中，列表的每个元素可以简单到一个元素，也可以复杂到一个矩阵、一个函数，甚至一个字符串向量。

### 创建和索引列表

在下面的例子中，我们将从开始，看看一些创建和初始化列表的常用方法。除此之外，我们还将看看如何访问这些列表元素来进行进一步的计算。请记住，列表中的每个元素都可以是简单的原始数据类型，甚至可以是复杂的数据结构或函数。

```r

> list.sample <- list(

+     1:5,

 +     c("first", "second", "third"),

+     c(TRUE, FALSE, TRUE, TRUE),

+     cos,

+     matrix(1:9, nrow = 3, ncol = 3)

+ )

> list.sample

```

**输出:**

![Creating and indexing lists](img/B03459_01_21.jpg)

```r

> list.with.names <- list(

+     even.nums = seq.int(2,10,2),

+     odd.nums  = seq.int(1,10,2),

+     languages = c("R", "Python", "Julia", "Java"),

+     cosine.func = cos

+ )

> list.with.names

```

**输出:**

![Creating and indexing lists](img/B03459_01_22.jpg)

```r

> list.with.names$cosine.func

function (x)  .Primitive("cos")

> list.with.names$cosine.func(pi)

[1] -1

>

> list.sample[[4]]

function (x)  .Primitive("cos")

> list.sample[[4]](pi)

[1] -1

>

> list.with.names$odd.nums

[1] 1 3 5 7 9

> list.sample[[1]]

[1] 1 2 3 4 5

> list.sample[[3]]

[1]  TRUE FALSE  TRUE  TRUE

```

从前面的例子中您可以看到访问列表中的任何元素并使用它进行进一步的计算是多么容易，例如`cos`函数。

### 合并和转换列表

现在，我们将在下面的示例中看看如何将几个列表组合成一个单个列表:

```r

> l1 <- list(

+     nums = 1:5,

+     chars = c("a", "b", "c", "d", "e"),

+     cosine = cos

+ )

> l2 <- list(

+     languages = c("R", "Python", "Java"),

+     months = c("Jan", "Feb", "Mar", "Apr"),

+     sine = sin

+ )

> # combining the lists now

> l3 <- c(l1, l2)

> l3

```

**输出:**

![Combining and converting lists](img/B03459_01_23.jpg)

很容易将列表转换成向量，反之亦然。下面的例子展示了我们实现这一目标的一些常用方法:

```r

> l1 <- 1:5

> class(l1)

[1] "integer"

> list.l1 <- as.list(l1)

> class(list.l1)

[1] "list"

> list.l1

```

**输出:**

![Combining and converting lists](img/B03459_01_24.jpg)

```r

> unlist(list.l1)

[1] 1 2 3 4 5

```

## 数据帧

数据框是一种特殊的数据结构，通常用于存储数据表或电子表格形式的数据，其中每一列表示一个特定的属性或字段，行由这些列的特定值组成。这种数据结构在处理通常具有大量字段和属性的数据集时非常有用。

### 创建数据帧

我们可以使用`data.frame`函数轻松创建数据帧。我们将看一些下面的例子来说明一些受欢迎的超级英雄也是如此:

```r

> df <- data.frame(

+     real.name = c("Bruce Wayne", "Clark Kent", "Slade Wilson", "Tony Stark", "Steve Rogers"),

+     superhero.name = c("Batman", "Superman", "Deathstroke", "Iron Man", "Capt. America"),

+     franchise = c("DC", "DC", "DC", "Marvel", "Marvel"),

+     team = c("JLA", "JLA", "Suicide Squad", "Avengers", "Avengers"),

+     origin.year = c(1939, 1938, 1980, 1963, 1941)

+ )

> df

```

**输出:**

![Creating data frames](img/B03459_01_25.jpg)

```r

> class(df)

[1] "data.frame"

> str(df)

```

**输出:**

![Creating data frames](img/B03459_01_26.jpg)

```r

> rownames(df)

[1] "1" "2" "3" "4" "5"

> colnames(df)

```

**输出:**

![Creating data frames](img/B03459_01_27.jpg)

```r

> dim(df)

[1] 5 5

```

`str`函数详细讲述了数据帧的结构，我们可以看到数据帧中每一列的数据细节。R base 中有很多现成的数据集，你可以直接加载并开始使用。其中一个如下所示。`mtcars`数据集包含各种汽车的信息，这些信息摘自 1974 年的*美国汽车趋势杂志*。

```r

> head(mtcars)   # one of the datasets readily available in R

```

**输出:**

![Creating data frames](img/B03459_01_28.jpg)

### 对数据帧进行操作

我们可以对数据帧进行很多操作，比如合并、组合、切片和转置数据帧。我们将在以下示例中查看一些重要的数据框操作。

使用简单的索引和函数(如`subset`)来索引和子集化数据框中的特定数据非常容易。

```r

> df[2:4,]

```

**输出:**

![Operating on data frames](img/B03459_01_29.jpg)

```r

> df[2:4, 1:2]

```

**输出:**

![Operating on data frames](img/B03459_01_30.jpg)

```r

> subset(df, team=="JLA", c(real.name, superhero.name, franchise))

```

**输出:**

![Operating on data frames](img/B03459_01_31.jpg)

```r

> subset(df, team %in% c("Avengers","Suicide Squad"), c(real.name, superhero.name, franchise))

```

**输出:**

![Operating on data frames](img/B03459_01_32.jpg)

我们现在来看看一些更复杂的操作，比如组合和合并数据帧。

```r

> df1 <- data.frame(

+     id = c('emp001', 'emp003', 'emp007'),

+     name = c('Harvey Dent', 'Dick Grayson', 'James Bond'),

+     alias = c('TwoFace', 'Nightwing', 'Agent 007')

+ )

> 

> df2 <- data.frame(

+     id = c('emp001', 'emp003', 'emp007'),

+     location = c('Gotham City', 'Gotham City', 'London'),

+     speciality = c('Split Persona', 'Expert Acrobat', 'Gadget Master')

+ )

> df1

```

**输出:**

![Operating on data frames](img/B03459_01_33.jpg)

```r

> df2

```

**输出:**

![Operating on data frames](img/B03459_01_34.jpg)

```r

> rbind(df1, df2)   # not possible since column names don't match

Error in match.names(clabs, names(xi)) : 

 names do not match previous names

> cbind(df1, df2)

```

**输出:**

![Operating on data frames](img/B03459_01_35.jpg)

```r

> merge(df1, df2, by="id")

```

**输出:**

![Operating on data frames](img/B03459_01_36.jpg)

从前面的操作可以明显看出，`rbind`和`cbind`的工作方式与我们之前看到的数组和矩阵的工作方式相同。但是，merge 允许您以连接关系数据库中各种表的相同方式合并数据框。



# 使用功能

接下来，我们将看看函数，这是一种技术或方法，可以轻松地构建和模块化您的代码，特别是执行特定任务的代码行，这样您就可以在需要时执行它们，而不必一次又一次地编写它们。在 R 中，函数基本上被视为另一种数据类型，您可以分配函数，在需要时操作它们，还可以将它们作为参数传递给其他函数。我们将在下一节探讨所有这些。

## 内置函数

R 由几个在 R-base 包中可用的函数组成，随着你安装更多的包，你会得到更多的功能，这些功能以函数的形式出现。我们将在以下示例中查看一些内置函数:

```r

> sqrt(5)

[1] 2.236068

> sqrt(c(1,2,3,4,5,6,7,8,9,10))

[1] 1.000000 1.414214 1.732051 2.000000 2.236068 2.449490 2.645751     [8] 2.828427 3.000000 3.162278

> # aggregating functions

> mean(c(1,2,3,4,5,6,7,8,9,10))

[1] 5.5

> median(c(1,2,3,4,5,6,7,8,9,10))

[1] 5.5

```

从前面的例子可以看出，`mean`、`median,`、`sqrt`等函数都是内置的，启动 R 就可以随时使用，不需要加载任何其他包，也不需要显式定义函数。

## 用户自定义功能

真正的强大之处在于能够根据您想要对数据执行的不同操作和计算来定义您自己的函数，并让 R 按照您想要的方式执行这些函数。一些插图如下所示:

```r

square <- function(data){

 return (data^2)

}

> square(5)

[1] 25

> square(c(1,2,3,4,5))

[1]  1  4  9 16 25

point <- function(xval, yval){

 return (c(x=xval,y=yval))

}

> p1 <- point(5,6)

> p2 <- point(2,3)

> 

> p1

x y 

5 6 

> p2

x y 

2 3

```

正如我们在前面的代码片段中看到的，我们可以定义像`square`这样的函数，它使用相同的代码计算单个数字的平方，甚至是一个数字向量。诸如 point 之类的函数对于在二维坐标空间中表示代表`points`的特定实体是有用的。现在我们将看看如何一起使用前面的函数。

## 将函数作为参数传递

当你定义一个函数时，如果你想在你的函数中使用其他函数来执行一些复杂的计算，你也可以把其他函数作为参数传递给它。这降低了代码的复杂性和冗余性。以下示例使用前面定义的`square`函数计算两点之间的欧几里德距离，该函数作为参数传递:

```r

> # defining the function

euclidean.distance <- function(point1, point2, square.func){

 distance <- sqrt(

 as.integer(

 square.func(point1['x'] - point2['x'])

 ) +

 as.integer(

 square.func(point1['y'] - point2['y'])

 )

 )

 return (c(distance=distance))

}

> # executing the function, passing square as argument

> euclidean.distance(point1 = p1, point2 = p2, square.func = square)

distance 

4.242641 

> euclidean.distance(point1 = p2, point2 = p1, square.func = square)

distance 

4.242641 

> euclidean.distance(point1 = point(10, 3), point2 = point(-4, 8), square.func = square)

distance

14.86607

```

因此，您可以看到，使用函数，您可以一次定义一个特定的函数，并根据需要多次执行它。



# 控制代码流

本节涵盖了与控制代码执行相关的领域。使用特定的构造，如`if-else`和`switch`，您可以有条件地执行代码。像`for`、`while`、`repeat`和`help`这样的构造在多次执行相同的代码时也称为循环。我们将在下一节探讨所有这些结构。

## 使用 if、if-else 和 ifelse

有几个构造帮助我们有条件地执行代码。当我们不想一个接一个地执行一堆语句，而是只在满足或不满足特定条件时执行代码时，这个特别有用。以下示例说明了这一点:

```r

> num = 5

> if (num == 5){

+     cat('The number was 5')

+ }

The number was 5

> 

> num = 7

> 

> if (num == 5){

+     cat('The number was 5')

+ } else{

+     cat('The number was not 5')

+ }

The number was not 5

>

> if (num == 5){

+     cat('The number was 5')

+ } else if (num == 7){

+     cat('The number was 7')

+ } else{

+     cat('No match found')

+ }

The number was 7

> ifelse(num == 5, "Number was 5", "Number was not 5")

[1] "Number was not 5"

```

## 使用开关工作

当您必须将一个表达式或参数与几个条件进行匹配，并且仅当存在特定匹配时才执行时，`switch` 函数特别有用。当用`if-else`构造实现时，这变得非常混乱，但用`switch`函数就优雅多了，我们接下来会看到:

```r

> switch(

+ "first",

+ first = "1st",

+ second = "2nd",

+ third = "3rd",

+ "No position"

+ )

[1] "1st"

> 

> switch(

+ "third",

+ first = "1st",

+ second = "2nd",

+ third = "3rd",

+ "No position"

+ )

[1] "3rd"

> # when no match, default statement executes

> switch(

+ "fifth",

+ first = "1st",

+ second = "2nd",

+ third = "3rd",

+ "No position"

+ )

[1] "No position"

```

## 循环

循环是在需要时重复执行代码段的一种极好的方式。然而，对于处理较大的数据集，矢量化构造比循环更加优化，但我们将在本章后面看到这一点。现在，你应该记住 R 中有三种类型的循环，即`for`、`while`和`repeat`。我们将在下面的示例中查看所有这些内容:

```r

> # for loop

> for (i in 1:10){

+     cat(paste(i," "))

+ }

1  2  3  4  5  6  7  8  9  10 

> 

> sum = 0

> for (i in 1:10){

+     sum <- sum + i

+ }

> sum

[1] 55

> 

> # while loop

> count <- 1

> while (count <= 10){

+     cat(paste(count, " "))

+     count <- count + 1

+ }

1  2  3  4  5  6  7  8  9  10 

> 

> # repeat infinite loop 

> count = 1

> repeat{

+     cat(paste(count, " "))

+     if (count >= 10){

+         break  # break off from the infinite loop

+     }

+     count <- count + 1

+ }

1  2  3  4  5  6  7  8  9  10 

```



# 高级构造

当我们谈论不使用循环对向量进行操作时，我们早些时候听到了术语向量化**。虽然循环是迭代向量和执行计算的一种很好的方式，但当我们处理所谓的**大数据**时，它不是很有效。在这种情况下，R 提供了一些高级的构造，我们将在本节中看到。我们将涵盖以下功能:**

*   `lapply`:循环遍历一个列表，并对每个元素计算一个函数
*   `sapply`:lapply 的简化版
*   `apply`:在数组的边界或边距上计算一个函数
*   `tapply`:对向量的子集计算函数
*   `mapply`:lapply 的多元版本

## 精力充沛

就像我们之前提到的一样，`lapply`将一个列表和一个函数作为输入，对列表中的每个元素计算该函数。如果输入列表不是列表，则在返回输出之前，使用`as.list`函数将其转换为列表。它比普通的循环快得多，因为实际的循环是使用 C 代码在内部完成的。我们在下面的代码片段中查看它的实现和示例:

```r

> # lapply function definition

> lapply

function (X, FUN, ...) 

{

 FUN <- match.fun(FUN)

 if (!is.vector(X) || is.object(X)) 

 X <- as.list(X)

 .Internal(lapply(X, FUN))

}

<bytecode: 0x00000000003e4f68>

<environment: namespace:base>

> # example

> nums <- list(l1=c(1,2,3,4,5,6,7,8,9,10), l2=1000:1020)

> lapply(nums, mean)

```

**输出:**

![lapply and sapply](img/B03459_01_37.jpg)

来到`sapply`，它与`lapply`相似，除了它试图尽可能简化结果。例如，如果最终结果是每个元素的长度都是`1`，则返回一个向量，如果结果中每个元素的长度都相同但大于 1，则返回一个矩阵，如果不能简化结果，则得到与`lapply`相同的结果。我们用下面的例子来说明这一点:

```r

> data <- list(l1=1:10, l2=runif(10), l3=rnorm(10,2))

> data

```

**输出:**

![lapply and sapply](img/B03459_01_38.jpg)

```r

> 

> lapply(data, mean)

```

**输出:**

![lapply and sapply](img/B03459_01_39.jpg)

```r

> sapply(data, mean)

```

**输出:**

![lapply and sapply](img/B03459_01_40.jpg)

## 申请

`apply`函数是,用于在数组的边缘或边界上计算函数；例如，对数组的行或列应用聚合函数。`rowSums`、`rowMeans`、`colSums,`和`colMeans`函数也在内部使用`apply`，但在大型数组上运行时会更加优化和有用。我们将在下面的示例中看到所有前面的构造:

```r

> mat <- matrix(rnorm(20), nrow=5, ncol=4)

> mat

```

**输出:**

![apply](img/B03459_01_41.jpg)

```r

> # row sums

> apply(mat, 1, sum)

[1]  0.79786959  0.53900665 -2.36486927 -1.28221227  0.06701519

> rowSums(mat)

[1]  0.79786959  0.53900665 -2.36486927 -1.28221227  0.06701519

> # row means

> apply(mat, 1, mean)

[1]  0.1994674  0.1347517 -0.5912173 -0.3205531  0.0167538

> rowMeans(mat)

[1]  0.1994674  0.1347517 -0.5912173 -0.3205531  0.0167538

>

> # col sums

> apply(mat, 2, sum)

[1] -0.6341087  0.3321890 -2.1345245  0.1932540

> colSums(mat)

[1] -0.6341087  0.3321890 -2.1345245  0.1932540

> apply(mat, 2, mean)

[1] -0.12682173  0.06643781 -0.42690489  0.03865079

> colMeans(mat)

[1] -0.12682173  0.06643781 -0.42690489  0.03865079

>

> # row quantiles

> apply(mat, 1, quantile, probs=c(0.25, 0.5, 0.75))

```

**输出:**

![apply](img/B03459_01_42.jpg)

因此，你可以看到在矩阵上应用各种统计函数是多么容易，根本不需要使用循环。

## 塔普利

函数`tapply`是,用于计算任何向量子集的函数。如果您熟悉使用关系数据库，这类似于在 SQL 中应用`GROUP BY`构造。我们在以下示例中对此进行了说明:

```r

> data <- c(1:10, rnorm(10,2), runif(10))

> data

```

**输出:**

![tapply](img/B03459_01_43.jpg)

```r

> groups <- gl(3,10)

> groups

 [1] 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 3 3 3 3 3 3

Levels: 1 2 3

> tapply(data, groups, mean)

```

**输出:**

![tapply](img/B03459_01_44.jpg)

```r

> tapply(data, groups, mean, simplify = FALSE)

```

**输出:**

![tapply](img/B03459_01_45.jpg)

```r

> tapply(data, groups, range)

```

**输出:**

![tapply](img/B03459_01_46.jpg)

## 地图应用

`mapply`函数是`lapply`的多元版本，用于在参数集上并行计算函数。一个简单的例子是，如果我们必须使用`rep`函数构建一个向量列表，我们必须多次编写它。然而，通过`mapply`，我们可以以一种更优雅的方式达到同样的效果，如下图所示:

```r

> list(rep(1,4), rep(2,3), rep(3,2), rep(4,1))

```

**输出:**

![mapply](img/B03459_01_47.jpg)

```r

> mapply(rep, 1:4, 4:1)

```

**输出:**

![mapply](img/B03459_01_47.jpg)



# 下一步用 R

在我们深入研究机器学习之前，暂停一会儿，深呼吸，思考一下到目前为止你学到了什么，这将是有益的。这篇快速而详细的 R 复习将在接下来的章节中对你有很大帮助。然而，还有两件事我们必须尽快处理。它们是如何在 R 中获得帮助以及如何使用 R 中的各种包。

## 寻求帮助

到目前为止，您一定已经发现 R 中有成千上万的函数和构造，不可能记住它们中的每一个实际上是做什么的，您也不需要记住！r 提供了许多直观的方法来获得关于任何函数、包或数据结构的帮助。首先，您可以在`R`命令提示符下运行`help.start()`函数，这将启动一个手动浏览器。在这里，您将获得有关 R 的详细信息，包括手册、参考资料和其他材料。以下命令显示了`help.start()`的内容，如命令后面的屏幕截图所示，您可以使用该命令进一步导航并获得更多帮助:

```r

> help.start()

```

如果没事，你自己开`http://127.0.0.1:31850/doc/html/index.html`。

![Getting help](img/B03459_01_48.jpg)

要获得 R 中任何特定函数或构造的帮助，如果您知道函数名，您可以使用帮助函数或结合函数名使用`?`操作符来获得帮助。例如，如果您需要关于`apply`功能的帮助，只需键入`help("apply")`或`?apply`即可获得关于`apply`功能的详细信息。这种在 R 中获得帮助的简单机制提高了您的生产率，并使 R 工作成为一种愉快的体验。通常，你不太记得你打算使用的函数的确切名称，但是你可能对它的名称有一个模糊的概念。r 也有一个帮助特性，你可以使用`help.search`函数或者`??`操作符，结合和函数名。例如，您可以使用`??apply`获得更多关于应用功能的信息。

## 搬运包裹

在**CRAN**(**Comprehensive R Archive Network**)上有成千上万个包含各种功能的包，这是一个托管所有这些包的存储库。要从 CRAN 下载任何包，你所要做的就是运行`install.packages`函数，将包名作为参数传递，就像`install.packages("caret")`一样。一旦下载并安装了这个包，您就可以使用`library`功能将它加载到您当前的 R 会话中。要加载包`caret`，只需输入`library(caret)`，它就可以使用了。`require`函数具有加载特定包的类似功能，并且以类似的方式在函数内部专门使用，通过键入`require(caret)`来加载`caret`包。`require`和`library`的唯一区别在于，如果找不到特定的包，`library`会显示错误，而`require`会继续执行代码而不显示任何错误。但是，如果有对该包的依赖调用，那么您的代码肯定会抛出一个错误。



# 机器学习基础知识

现在你已经刷新了关于 R 的记忆，我们将讨论什么是机器学习的基础知识，它现在是如何使用的，以及机器学习的主要领域。本节旨在提供机器学习的概述，这将有助于为下一章铺平道路，在下一章我们将更深入地探索它。

## 机器学习——它到底意味着什么？

机器学习不只有一个独特的教科书定义，因为它是一个包含并借鉴了计算机科学其他几个领域的概念和技术的领域。它也是大学里的一门学术课程，最近变得更加突出，机器学习和数据科学以教育视频、课程和培训的形式在网上被广泛采用。机器学习基本上是来自计算机科学、统计学和数学领域的元素的交叉，它使用来自人工智能、模式检测、优化和学习理论的概念来开发算法和技术，这些算法和技术可以从数据中学习并对数据进行预测，而无需显式编程。

这里的学习是指让计算机或机器基于我们提供给它们的数据和算法变得智能的能力，以便它们开始从提供的数据中检测模式和洞察力。这种学习确保机器可以检测输入给它的数据的模式，而不必每次都显式地对它们进行编程。初始数据或观察值被输入到机器中，机器学习算法对这些数据进行处理，以生成一些输出，这些输出可以是预测、假设，甚至是一些数值结果。基于这个输出，可以有对我们的机器学习算法的反馈机制来改进我们的结果。这整个系统形成了一个机器学习模型，它可以直接用于全新的数据或观察，以从中获得结果，而无需再次编写任何单独的算法来处理这些数据。

## 机器学习——它在世界上是如何使用的？

你可能想知道一些算法或代码在现实世界中究竟是如何使用的。事实证明，它们在不同垂直行业的各种用例中都有使用。一些例子如下:

*   **零售**:机器学习广泛应用于零售和电子商务垂直领域，每个商店都希望超越竞争对手。

    *   **定价分析**:机器学习算法用于比较不同商店的商品价格，以便商店能够以最具竞争力的价格出售商品。
    *   **购物篮分析**:用于分析客户购物趋势，推荐购买产品，我们将在[第三章](ch03.html "Chapter 3. Predicting Customer Shopping Trends with Market Basket Analysis")、*用购物篮分析预测客户购物趋势*中介绍。
    *   **推荐引擎**:用于分析客户购买、评分、满意度，向各类用户推荐产品。我们将在[第四章](ch04.html "Chapter 4. Building a Product Recommendation System")、*建立一个产品推荐系统*中建立一些我们自己的推荐系统。

*   **广告**:广告行业严重依赖机器学习，向消费者推广和展示正确的广告，以获得最大的转化。

    *   **网络分析**:分析网站流量
    *   **客户流失率分析**:预测客户流失率
    *   **广告点击率预测**:用于预测广告对消费者的有效程度，以便消费者点击广告购买相关产品

*   **医疗保健**:机器学习算法被广泛应用于医疗保健行业，以更有效地治疗患者。

    *   **疾病检测和预测**:用于根据患者的病史检测和预测患病几率。
    *   研究人脑和 DNA 等复杂结构，以更好地了解人体的功能，进行更有效的治疗。

*   垃圾邮件和消息的检测和过滤。
*   预测选举结果。
*   欺诈检测和预测。我们将在[第 5 章](ch05.html "Chapter 5. Credit Risk Detection and Prediction – Descriptive Analytics")、*信用风险检测和预测-描述性分析*和[第 6 章](ch06.html "Chapter 6. Credit Risk Detection and Prediction – Predictive Analytics")、*信用风险检测和预测-预测性分析*中尝试解决最关键的欺诈检测问题之一。
*   消息应用程序中的文本预测。
*   无人驾驶汽车、飞机和其他交通工具。
*   天气、交通和犯罪活动的预测和预报。
*   情绪和情感分析，我们将在[第 8 章](ch08.html "Chapter 8. Sentiment Analysis of Twitter Data")、*Twitter 数据的情绪分析*中介绍。

前面的例子只是触及了机器学习真正能做什么的表面，到目前为止，我确信您已经对机器学习被广泛使用的各个领域有了很好的了解。

## 机器学习算法的类型

正如我们之前谈到的，要让机器学习，你需要机器学习算法。机器学习算法是一类特殊的算法，它处理数据并从中收集见解。这个想法是使用数据和算法的组合来建立一个模型，然后可以用来处理新数据并获得可操作的见解。

每种机器学习算法都取决于它可以处理什么类型的数据，以及我们试图解决什么类型的问题。你可能很想学习一些算法，然后尝试将它们应用到你面临的每个问题中。请记住，没有适合所有问题的通用机器学习算法。机器学习算法的主要输入是由特征组成的数据，其中每个特征都可以描述为数据集的一个属性，例如你的身高、体重等，如果我们处理的是与人类相关的数据。机器学习算法可以分为两个主要领域，即监督和非监督学习算法。

### 监督机器学习算法

受监督的学习算法是主要用于预测建模的机器学习算法家族的子集。预测模型基本上是由机器学习算法和来自训练数据的特征或属性构建的模型，这样我们可以使用从输入数据获得的其他值来预测一个值。监督学习算法尝试对目标预测输出和输入要素之间的关系和依赖关系进行建模，以便我们可以根据从以前的数据集中学习到的关系来预测新数据的输出值。监督学习算法的主要类型包括:

*   **分类算法**:这些算法从具有特征和类别标签的训练数据中建立预测模型。这些预测模型反过来使用从新的、以前未见过的数据的训练数据中学习的特征来预测它们的类别标签。输出类是离散的。分类算法的类型包括决策树、随机森林、支持向量机等等。我们将在[第 2 章](ch02.html "Chapter 2. Let's Help Machines Learn")、*中使用其中的几种算法，让我们帮助机器学习*、[第 6 章](ch06.html "Chapter 6. Credit Risk Detection and Prediction – Predictive Analytics")、*信用风险检测和预测——预测分析*，以及[第 8 章](ch08.html "Chapter 8. Sentiment Analysis of Twitter Data")、*推特数据的情绪分析*。
*   **回归算法**:这些算法用于根据从数据中获得的一些输入特征来预测输出值。为此，该算法基于训练数据的特征和输出值建立模型，并且该模型用于预测新数据的值。这种情况下的输出值是连续的，而不是离散的。回归算法的类型包括线性回归、多元回归、回归树和套索回归等。我们在第二章、*让我们帮助机器学习*中探索其中的一些。

### 无监督机器学习算法

无监督学习算法是机器学习算法家族，主要用于模式检测和描述性建模。描述性模型基本上是由无监督的机器学习算法和来自输入数据的特征构建的模型，类似于监督学习过程。但是，这里没有输出类别或标签，算法可以基于这些类别或标签来尝试对关系进行建模。这些算法尝试对输入数据使用技术来挖掘规则、检测模式，并对数据点进行总结和分组，这有助于获得有意义的见解，并向用户更好地描述数据。这里没有训练或测试数据的具体概念，因为我们没有任何具体的关系映射，我们只是试图从我们试图分析的数据中获得有用的见解和描述。无监督学习算法的主要类型包括:

*   **聚类算法**:这些算法的主要目标是仅使用从输入数据单独导出的特征，而不使用其他外部信息，将输入数据点聚类或分组到不同的类或类别中。与分类不同，聚类中的输出标签是未知的。有不同的方法来建立聚类模型，例如使用均值、中间值、层次结构等等。一些流行的聚类算法包括 k-means、k-medoids 和层次聚类。我们将在[第二章](ch02.html "Chapter 2. Let's Help Machines Learn")、*让我们帮助机器学习*和[第七章](ch07.html "Chapter 7. Social Media Analysis – Analyzing Twitter Data")、*社交媒体分析——分析 Twitter 数据*中了解一些聚类算法。
*   **关联规则学习算法**:这些算法用于从数据集中挖掘和提取规则和模式。这些规则解释了不同变量和属性之间的关系，也描述了数据中出现的频繁项目集和模式。这些规则反过来有助于从庞大的数据仓库中发现对任何企业或组织都有用的洞察力。流行的算法有 Apriori 和 FP Growth。我们将在[第 2 章](ch02.html "Chapter 2. Let's Help Machines Learn")、*让我们帮助机器学习*和[第 3 章](ch03.html "Chapter 3. Predicting Customer Shopping Trends with Market Basket Analysis")、*使用购物篮分析预测顾客购物趋势*中使用其中一些。

### R 中流行的机器学习包

在简要了解了机器学习的基础知识和算法类型之后，你一定会好奇我们如何使用 R 来应用这些算法来解决现实世界的问题。事实证明，R 中有很多专门解决机器学习问题的包。这些软件包由算法组成，这些算法已经过优化，可以用来解决问题。我们将在 R 中列出几个流行的机器学习包，这样你就知道你以后可能需要什么工具，并且在后面的章节中使用这些包的时候会感觉更熟悉。根据用途和功能，以下 R 包在解决机器学习问题时非常受欢迎:

*   这个包(分类和回归训练的缩写)由几个机器学习算法组成，用于建立预测模型
*   这个包处理分类和回归的随机森林算法的实现
*   `rpart`:这个包关注递归分割和决策树
*   `glmnet`:这个包的主要焦点是 lasso 和弹性网正则化回归模型
*   `e1071`:它处理傅立叶变换，聚类，支持向量机，以及更多的监督和非监督算法
*   `party`:这处理递归分割
*   `arules`:这个包用于关联规则学习算法
*   这是一个用来构建推荐引擎的库
*   `nnet`:该软件包支持使用神经网络进行预测建模
*   `h2o`:这是目前数据科学中最受欢迎的软件包之一，提供了快速和可扩展的算法，包括梯度推进和深度学习

除了前面的库，还有很多其他与 r 中的机器学习相关的包。重要的是根据手头的数据和问题选择正确的算法和模型。



# 摘要

在这一章中，我们简要地谈论了我们将进入机器学习和 R 世界的旅程。我们讨论了 R 的基础知识，并为 R 中使用的核心构造和数据结构打下了坚实的基础。然后，我们通过查看一些概念和算法，以及如何在世界上使用它来解决问题，深入到机器学习的世界中。最后，我们快速浏览了一下 R 中一些最流行的机器学习包，让我们都熟悉了一些机器学习工具箱的便利工具！

在下一章，我们将深入研究机器学习的概念和算法，它们将帮助我们让机器学习一些东西！