

# 零、前言

当谈到轻松执行计算统计(统计计算)和探索机器学习的数学方面时，r 是最受欢迎的语言之一。通过这本书，你将利用 R 生态系统来构建高效的机器学习应用程序，在你的组织内执行智能任务。

这本书将帮助你测试你的知识和技能，指导你如何建立从简单到复杂的机器学习项目。首先，您将学习如何构建强大的机器学习模型来预测员工流失。接下来，您将实现一个笑话推荐引擎来对 Amazon 评论执行情感分析。您还将探索使用批发数据对客户进行细分的不同聚类技术。除此之外，这本书将让你熟悉使用自编码器的信用卡欺诈检测，以及强化学习，以做出预测并在赌场老虎机上获胜。

到本书结束时，您将能够自信地执行复杂的任务，为自动化操作构建研究和商业项目。



# 这本书是给谁的

这本书是为那些希望通过构建真实世界的项目来掌握使用 R 的 ML 概念的数据分析师、数据科学家和 ML 开发者而写的。每个项目将帮助你测试你的专业知识，以实现最大似然算法和技术的工作机制。对 ML 的基本理解和 R 编程的工作知识是必须的。



# 这本书涵盖的内容

[第 1 章](d8e2df34-05df-451e-88ce-62fdf17184d4.xhtml)，*探索机器学习的前景*，将简要回顾从业者必须知道的各种 ML 概念。在这一章中，我们将涵盖监督学习、强化学习、非监督学习和真实世界的 ML 用例等主题。

[第 2 章](664c4f7e-187f-4fa0-b796-af7db785bb08.xhtml)，*使用集成模型预测员工流失*，涵盖了通过集成学习创建强大的 ML 模型。本章涉及的项目来自人力资源领域。留住有才华的员工是企业面临的一个关键挑战。如果我们能够提前预测员工的流失，人力资源或管理团队就有可能采取一些措施来避免潜在的流失成为现实。碰巧的是，有可能通过应用 ML 来预测员工流失。这一章利用了 IBM 策划的公共数据集，该数据集提供了一个虚拟的员工流失人口和特征。我们以手头问题的介绍开始这一章，然后尝试用**探索性数据分析** ( **EDA** )探索数据集。下一步是预处理阶段，包括使用先前的领域经验创建新的特性。一旦数据集完全准备好，将使用多种集成技术创建模型，如装袋、提升、堆叠和随机化。最后，我们将为生产部署最终选择的模型。我们还将了解用于创建模型的各种集成技术的基本概念。

[第 3 章](9f39b0b1-736a-4b0d-9ca9-276096fe2ada.xhtml)，*实现笑话推荐引擎*，介绍推荐引擎，它被设计用来预测用户对电影和音乐等内容的评分。基于用户先前喜欢或看过的内容并使用其他简档属性，推荐引擎建议用户可能喜欢的新内容。近年来，这种发动机获得了很大的重要性。我们通过致力于一个笑话推荐引擎项目来探索令人兴奋的推荐系统领域。在这一章中，我们从理解协同过滤算法的概念和类型开始。然后，我们将构建一个推荐引擎，使用协作过滤方法(如基于用户的协作过滤器和基于项目的协作过滤器)来提供个性化的笑话推荐。这个项目使用的数据集是一个开放的数据集，称为 Jester 笑话数据集。除此之外，我们将探索 R 中可用于构建推荐系统的各种库，并且我们将比较从这些方法获得的性能。此外，我们利用市场购物篮分析技术，一种在营销领域非常流行的技术，来辨别各种笑话之间的关系。

[第 4 章](dd2c7d12-7674-49e3-8965-af815f5c6f07.xhtml)，*用 NLP* 对亚马逊评论进行情感分析，涵盖了情感分析，这需要找到一个句子的情感，并将其标记为正面、负面或中性。本章介绍了情感分析，并涵盖了可以用来分析文本的各种技术。我们将理解文本挖掘的概念，以及基于语气标记文本的各种方式。

我们将对亚马逊产品评论数据进行情感分析。该数据集包含数百万条亚马逊客户评论和星级评分。这是一个分类任务，我们将根据语气将每个评论分为积极、消极或中立。除了使用各种流行的 R 文本挖掘库来预处理要分类的评论，我们还将利用各种各样的文本表示，如 bag of words、word2vec、fastText 和 Glove。每个文本表示然后被用作 ML 算法的输入来执行分类。在实现这些技术的过程中，我们还将了解这些技术背后的概念，并探索我们可以成功应用它们的其他实例。

[第 5 章](43de56cf-5bfa-44c0-be57-b20df236cc95.xhtml)，*利用批发数据进行客户细分*，涵盖了客户的细分、分组或聚类，这可以通过无监督学习来实现。在本章中，我们将探讨客户分组的各个方面。客户细分是产品销售者用来了解客户和收集信息的重要工具。可以根据不同的标准对客户进行细分，例如年龄和消费模式。在这一章中，我们将学习各种客户细分技术。对于这个项目，我们使用一个包含批发交易的数据集。该数据集可在 UCI 机器学习库中获得。我们将应用先进的聚类技术，如 k-means，DIANA 和 AGNES。有时，我们不知道现有数据集中存在的组的数量。我们将探索处理这种模糊性的 ML 技术，并让 ML 根据输入数据的潜在特征找出可能的组数。评估聚类算法的输出对于从业者来说是一个经常具有挑战性的领域。我们也探索这一领域，以便对将聚类算法应用于现实世界的问题有一个全面的理解。

[第六章](018d53f7-b892-44e8-b45f-4d9041f9d65b.xhtml)，*利用深度神经网络*进行图像识别，涵盖了 **c** **选择性神经网络**(**CNN**)，这是一种深度神经网络，在计算机视觉应用中很流行。在这一章中，我们学习 CNN 的基本概念。我们探索了为什么细胞神经网络在处理计算机视觉问题，如物体检测时如此有效。我们讨论了迁移学习的各个方面，以及它如何与 CNN 协同工作来解决计算机视觉问题。正如书中其他地方一样，我们将遵循边做边学的哲学。我们将通过应用 CNN 在一个叫做 MNIST 的流行开放数据集上构建多类分类模型来学习所有这些概念。该项目的目标是对给定的手写数字图像进行分类。该项目探索了从原始图像创建特征的方法。我们将了解可应用于图像数据的各种预处理技术，以便将数据用于深度学习模型。

[第 7 章](002b0fa6-1419-42a8-bdb0-dd4bc422489a.xhtml)，*使用自编码器的信用卡欺诈检测*，涵盖了自编码器，这是另一种类型的无监督深度学习网络。我们通过理解自编码器以及它们与其他深度学习网络(如**循环神经网络** ( **RNNs** )和 CNN)的不同来开始这一章。我们将通过实施识别信用卡欺诈的项目来了解自编码器。信用卡公司一直在寻找检测信用卡欺诈的方法。欺诈检测是银行保护收入的一个重要方面。这可以通过在金融领域中针对特定的欺诈检测问题应用 ML 来实现。欺诈通常是需要立即采取行动的异常事件。在本章中，我们将使用自编码器来检测欺诈。自编码器是包含瓶颈层的神经网络，其维度小于输入数据。在这一章中，我们将熟悉降维，以及如何使用它来识别信用卡欺诈检测。对于该项目，我们将与 r 一起使用 H2O 深度学习框架。就数据集而言，我们使用一个开放的数据集，其中包含 2013 年 9 月以来欧洲持卡人的信用卡交易。总共有 284，807 笔交易，其中 492 笔是欺诈性的。

[第八章](265e576c-694a-476f-be7a-72281b2fc662.xhtml)、*用循环神经网络*自动生成散文，介绍一些最近受到很多关注的**深度神经网络** ( **DNNs** )。这是由于他们在 ML 的各个领域取得了巨大的成功，从人脸识别和物体检测到音乐生成和神经艺术，本章介绍了理解深度学习所必需的概念。我们讨论神经网络的具体细节，如神经元、隐藏层、各种激活函数、处理神经网络中面临的问题的技术，以及使用优化算法来获得神经网络中的权重。我们还将从头实现一个神经网络来演示这些概念。本章的内容将帮助我们获得神经网络的基础知识。然后，我们将通过做一个项目来学习如何应用 RNN。人们一直认为创作故事、写诗、画图等创造性任务只能由人类来完成。这已经不是事实了，感谢深度学习！技术现在可以完成创造性的任务。我们将创建一个基于**长短期记忆** ( **LSTM** )网络的应用程序，它是自动生成文本的 RNNs 的变体。为了完成这项任务，我们利用 MXNet 框架，它扩展了对 R 语言的支持，以执行深度学习。在实施这个项目的过程中，我们还将了解更多关于 rnn 和 LSTMs 的概念。

[第 9 章](4b80233e-4fbe-4d90-ba32-5053930433c1.xhtml)、*通过强化学习赢得赌场吃角子老虎机*，从 RL 的解释开始。我们讨论 RL 的各种概念，包括解决所谓的多臂老虎机问题的策略。我们实现了一个项目，使用 UCB 和汤普森采样技术，以解决多臂老虎机问题。

[附录](d17f7a4d-ea3f-4d43-bc59-0cf260a8ac49.xhtml)、*未来之路*，简要讨论了 ML 世界的进步以及保持领先的必要性。



# 从这本书中获得最大收益

本书涵盖的项目旨在让你接触到在现实世界问题中实现各种 ML 技术的实用知识。期望你对 R 有很好的工作知识，并对 ML 有一些基本的了解。在开始这个项目之前，ML 和 R 的基本知识是必须的。

还应该注意的是，项目的代码是使用 R 版本 3.5.2 (2018-12-20)实现的，昵称为蛋壳冰屋。项目代码已经在 Linux Mint 18.3 Sylvia 上测试成功。没有理由认为代码在其他平台上不起作用，比如 Windows 但是，这不是笔者验证过的东西。



# 下载示例代码文件

你可以从你在[www.packt.com](http://www.packt.com)的账户下载本书的示例代码文件。如果你在其他地方购买了这本书，你可以访问 www.packt.com/support 的[并注册，让文件直接通过电子邮件发送给你。](http://www.packt.com/support)

您可以按照以下步骤下载代码文件:

1.  在 www.packt.com[登录或注册](http://www.packt.com)。
2.  选择支持选项卡。
3.  点击代码下载和勘误表。
4.  在搜索框中输入图书名称，然后按照屏幕指示进行操作。

下载文件后，请确保使用最新版本的解压缩或解压文件夹:

*   WinRAR/7-Zip for Windows
*   适用于 Mac 的 Zipeg/iZip/UnRarX
*   用于 Linux 的 7-Zip/PeaZip

这本书的代码包也托管在 GitHub 的 https://GitHub . com/packt publishing/R-Machine-Learning-Projects 上。如果代码有更新，它将在现有的 GitHub 库中更新。

我们在 https://github.com/PacktPublishing/也有来自我们丰富的书籍和视频目录的其他代码包。看看他们！



# 下载彩色图像

我们还提供了一个 PDF 文件，其中有本书中使用的截图/图表的彩色图像。可以在这里下载:[http://www . packtpub . com/sites/default/files/downloads/9781789807943 _ color images . pdf](http://www.packtpub.com/sites/default/files/downloads/Bookname_ColorImages.pdf)。



# 使用的惯例

本书通篇使用了许多文本约定。

`CodeInText`:表示文本中的码字、数据库表名、文件夹名、文件名、文件扩展名、路径名、伪 URL、用户输入和 Twitter 句柄。下面是一个例子:“`rsample`库包含了这个数据集，我们可以直接从库中使用这个数据集。”

代码块设置如下:

```r
setwd("~/Desktop/chapter 2") 
library(rsample)
data(attrition) 
str(attrition) 
mydata<-attrition 
```

当我们希望将您的注意力吸引到代码块的特定部分时，相关的行或项目以粗体显示:

```r
[default]
exten => s,1,Dial(Zap/1|30)
exten => s,2,Voicemail(u100)
exten => s,102,Voicemail(b100)
exten => i,1,Voicemail(s0)
```

**粗体**:表示一个新术语、一个重要的单词或您在屏幕上看到的单词。例如，菜单或对话框中的单词出现在文本中，如下所示。这里有一个例子:“你可能还记得买了这个商品的顾客也在亚马逊(或任何电子商务网站)上买了这个标题，那里有推荐。”

警告或重要提示如下所示。

提示和技巧是这样出现的。



# 取得联系

我们随时欢迎读者的反馈。

总体反馈:如果你对这本书的任何方面有疑问，请在邮件主题中提及书名，并发邮件至`customercare@packtpub.com`。

**勘误表**:虽然我们已经尽力确保内容的准确性，但错误还是会发生。如果你在这本书里发现了一个错误，请告诉我们，我们将不胜感激。请访问[www.packt.com/submit-errata](http://www.packt.com/submit-errata)，选择您的图书，点击勘误表提交表单链接，并输入详细信息。

**盗版**:如果您在互联网上遇到我们作品的任何形式的非法拷贝，如果您能提供我们的地址或网站名称，我们将不胜感激。请通过`copyright@packt.com`联系我们，并提供材料链接。

**如果你有兴趣成为一名作家**:如果有一个你擅长的主题，并且你有兴趣写书或投稿，请访问[authors.packtpub.com](http://authors.packtpub.com/)。



# 复习

请留下评论。一旦你阅读并使用了这本书，为什么不在你购买它的网站上留下评论呢？潜在的读者可以看到并使用您不带偏见的意见来做出购买决定，我们 Packt 可以了解您对我们产品的看法，我们的作者可以看到您对他们的书的反馈。谢谢大家！

更多关于 Packt 的信息，请访问[packt.com](http://www.packt.com/)。