

# 零、前言

机器学习通过将数据转化为可操作的情报，对学术界和工业界产生了巨大的影响。另一方面，Scala 在过去几年中的应用稳步增长，尤其是在数据科学和分析领域。这本书是为数据科学家、数据工程师和深度学习爱好者编写的，他们具有复杂数值计算的坚实背景，并希望学习更多的动手机器学习应用程序开发。

因此，如果你精通机器学习概念，并且想要通过使用 Scala 的力量深入实际实现来扩展你的知识，那么这本书就是你所需要的！通过 11 个端到端的项目，您将熟悉流行的机器学习库，如 Spark ML、H2O、Zeppelin、DeepLearning4j 和 MXNet。

在阅读了这本书并实践了所有的项目之后，你将能够支配数值计算、深度学习和函数式编程来执行复杂的数值任务。因此，您可以在生产就绪环境中开发、构建和部署研究和商业项目。

这本书不应该从头到尾读一遍。你可以翻到某一章，看起来像是你正在努力完成的事情，或者只是点燃你的兴趣。但是任何改进反馈都是受欢迎的。

快乐阅读！



# 这本书是给谁的

如果你想利用 Scala 和 Spark ML、Deeplearning4j、H2O、MXNet 和 Zeppelin 等开源库的力量来理解大数据，那么这本书就是为你准备的。建议对 Scala 和 Scala Play 框架有深刻的理解。熟悉 ML 技术将是一个额外的优势。



# 这本书涵盖的内容

[第 1 章](4b0be2d2-f313-471f-83fe-830931fc8af9.xhtml)，*分析保险严重索赔*，展示了如何使用一些广泛使用的回归技术开发一个分析保险严重索赔的预测模型。我们将演示如何在生产就绪环境中部署该模型。

[第 2 章](4e196881-40c8-4eb9-b2b3-e332a49adc1a.xhtml)、*分析和预测电信客户流失*，使用 Orange 电信客户流失数据集，该数据集由经过清理的客户活动和客户流失标签组成，用于指定客户是否取消订购，以开发现实生活中的预测模型。

[第 3 章](51e66c26-e12b-4764-bbb7-444986c05870.xhtml)，*根据历史和实时数据进行高频比特币价格预测*，展示了如何开发一个收集历史和实时数据的真实项目。我们预测未来几周、几个月的比特币价格，等等。此外，我们还演示了如何为比特币在线交易生成一个简单的信号。最后，本章使用 Scala Play 框架将整个应用程序包装成一个 web 应用程序。

[第 4 章](249a2b25-f72b-417f-961c-6f3d5ab8acce.xhtml)，*人口规模聚类和种族预测*，使用来自 1000 基因组计划的基因组变异数据，将 K-means 聚类方法应用于可扩展的基因组数据分析。这是为了在群体规模上聚类基因型变异。最后，我们训练深度神经网络和随机森林模型来预测种族。

[第 5 章](9e9b701c-32ce-4a11-9007-68dc11f8ce69.xhtml)，*自然语言处理中的主题建模——更好地洞察大规模文本*，展示了如何利用基于 Spark 的 LDA 算法和斯坦福自然语言处理大规模原始文本来开发主题建模应用。

[第 6 章](c1167c6a-b128-4eca-b085-0532b126c4aa.xhtml)，*开发* *基于模型的电影推荐引擎*，展示了如何通过奇异值分解、ALS 和矩阵分解之间的互操作来开发可扩展的电影推荐引擎。电影镜头数据集将用于这个端到端项目。

[第 7 章](c4a322da-d64b-4c40-a5b8-7ffec8381b41.xhtml)，*使用 Q-Learning 和 Scala Play 框架的期权交易*，在真实的 IBM 股票数据集上应用强化 QLearning 算法，并设计一个由批评和奖励驱动的机器学习系统。目标是开发一个名为**期权交易**的现实应用。本章使用 Scala Play 框架将整个应用程序包装成一个 web 应用程序。

[第 8 章](3e09dbd3-a9bb-4451-97f1-1a961d28b4a0.xhtml)、*使用深度神经网络的银行电话营销客户订阅评估*，是一个端到端的项目，展示了如何解决一个名为**客户订阅评估**的现实问题。将使用银行电话销售数据集来训练 H2O 深度神经网络。最后，本章对该预测模型的性能进行了评估。

[第 9 章](d10834fe-9b85-4ecd-a818-6c223b6c435d.xhtml)、*使用自编码器和异常检测的欺诈分析*，使用自编码器和异常检测技术进行欺诈分析。使用的数据集是在 Worldline 和 ULB 大学的机器学习小组的研究合作中收集和分析的欺诈检测数据集。

[第 10 章](ef398231-89c4-48f9-861b-ca50663aa455.xhtml)，*使用循环神经网络进行人体活动识别*，包括另一个端到端项目，展示了如何使用一个名为 LSTM 的 RNN 实现，通过智能手机传感器数据集进行人体活动识别。

[第 11 章](de2a960c-4b75-4ff0-ae95-a7d54be2e7dd.xhtml)，*使用卷积神经网络进行图像分类*，演示了如何开发预测分析应用程序，如图像分类，在名为 Yelp 的真实图像数据集上使用卷积神经网络。



# 从这本书中获得最大收益

这本书是献给开发人员、数据分析师和深度学习爱好者的，他们没有太多复杂数值计算的背景，但想知道什么是深度学习。建议对 Scala 及其函数式编程概念有深刻的理解。为了理解这本书，对 Spark ML、H2O、Zeppelin、DeepLearning4j 和 MXNet 的一些基本理解和高级知识将是一个额外的优势。此外，还假设了诸如 Maven 和 SBT 等构建工具的基本知识。

所有的例子都是用 Scala 在 Ubuntu 16.04 LTs 64 位和 Windows 10 64 位上实现的。您还需要以下内容(最好是最新版本):

*   Apache Spark 2.0.0(或更高版本)
*   MXNet、Zeppelin、DeepLearning4j 和 H2O(详见本章和提供的`pom.xml`文件)
*   Hadoop 2.7(或更高版本)
*   Java (JDK 和 JRE) 1.7+/1.8+版本
*   Scala 2.11.x(或更高版本)
*   带有 Maven 插件(2.9+)、Maven 编译器插件(2.3.2+)和 Maven 汇编插件(2.4.1+)的 Eclipse Mars 或 Luna(最新)
*   智能集成电路
*   安装了 SBT 插件和 Scala Play 框架

需要至少配备酷睿 i3 处理器、酷睿 i5(推荐)或酷睿 i7(以获得最佳结果)的计算机。然而，多核处理将提供更快的数据处理和可扩展性。对于独立模式，建议至少 8gb RAM；对于单个虚拟机，至少使用 32 GB RAM，对于群集，使用更高的 RAM。您应该有足够的存储来运行繁重的作业(取决于您将处理的数据集大小)；最好至少有 50 GB 的空闲磁盘存储空间(对于独立数据库和 SQL Warehouse)。

Linux 发行版是首选(包括 Debian、Ubuntu、Fedora、RHEL、CentOS 等等)。更具体地说，例如，对于 Ubuntu，建议使用 14.04 (LTS) 64 位(或更高版本)完整安装、VMWare player 12 或 VirtualBox。您可以在 Windows (XP/7/8/10)或 Mac OS X (10.4.7 以上)上运行 Spark 作业。



# 下载示例代码文件

你可以从你在[www.packtpub.com](http://www.packtpub.com)的账户下载本书的示例代码文件。如果你在别处购买了这本书，你可以访问 www.packtpub.com/support 并注册，让文件直接通过电子邮件发送给你。

您可以按照以下步骤下载代码文件:

1.  在[www.packtpub.com](http://www.packtpub.com/support)登录或注册。
2.  选择支持选项卡。
3.  点击代码下载和勘误表。
4.  在搜索框中输入图书名称，然后按照屏幕指示进行操作。

下载文件后，请确保使用最新版本的解压缩或解压文件夹:

*   WinRAR/7-Zip for Windows
*   适用于 Mac 的 Zipeg/iZip/UnRarX
*   用于 Linux 的 7-Zip/PeaZip

这本书的代码包也托管在 GitHub 的 https://GitHub . com/packt publishing/Scala-Machine-Learning-Projects 上。我们在 https://github.com/PacktPublishing/也有来自我们丰富的书籍和视频目录的其他代码包。看看他们！



# 下载彩色图像

我们还提供了一个 PDF 文件，其中有本书中使用的截图/图表的彩色图像。可以在这里下载:[http://www . packtpub . com/sites/default/files/downloads/scalamachinelementprojects _ color images . pdf](http://www.packtpub.com/sites/default/files/downloads/ScalaMachineLearningProjects_ColorImages.pdf)。



# 使用的惯例

本书通篇使用了许多文本约定。

`CodeInText`:表示文本中的码字、数据库表名、文件夹名、文件名、文件扩展名、路径名、伪 URL、用户输入和 Twitter 句柄。下面是一个例子:“将下载的`WebStorm-10*.dmg`磁盘镜像文件挂载为系统中的另一个磁盘。”

代码块设置如下:

```scala
val cv = new CrossValidator()
      .setEstimator(pipeline)
      .setEvaluator(new RegressionEvaluator)
      .setEstimatorParamMaps(paramGrid)
      .setNumFolds(numFolds)
```

Scala 函数代码块如下所示:

```scala
 def variantId(genotype: Genotype): String = {
      val name = genotype.getVariant.getContigName
      val start = genotype.getVariant.getStart
      val end = genotype.getVariant.getEnd
      s"$name:$start:$end"
  }
```

当我们希望将您的注意力吸引到代码块的特定部分时，相关的行或项目以粗体显示:

```scala
var paramGrid = new ParamGridBuilder()
      .addGrid(dTree.impurity, "gini" :: "entropy" :: Nil)
      .addGrid(dTree.maxBins, 3 :: 5 :: 9 :: 15 :: 23 :: 31 :: Nil)
      .addGrid(dTree.maxDepth, 5 :: 10 :: 15 :: 20 :: 25 :: 30 :: Nil)
      .build()
```

任何命令行输入或输出都按如下方式编写:

```scala
$ sudo mkdir Bitcoin
$ cd Bitcoin
```

**粗体**:表示新术语、重要单词或您在屏幕上看到的单词。例如，菜单或对话框中的单词出现在文本中，如下所示。下面是一个例子:“从管理面板中选择系统信息。”

警告或重要提示如下所示。

提示和技巧是这样出现的。



# 取得联系

我们随时欢迎读者的反馈。

**总体反馈**:给`feedback@packtpub.com`发邮件，在邮件主题中提及书名。如果您对本书的任何方面有任何疑问，请通过`questions@packtpub.com`发邮件给我们。

**勘误表**:虽然我们已经尽力确保内容的准确性，但错误还是会发生。如果你在这本书里发现了一个错误，请告诉我们，我们将不胜感激。请访问[www.packtpub.com/submit-errata](http://www.packtpub.com/submit-errata)，选择您的图书，点击勘误表提交表格链接，并输入详细信息。

**盗版**:如果你在互联网上遇到任何形式的非法拷贝，如果你能提供地址或网站名称，我们将不胜感激。请通过`copyright@packtpub.com`联系我们，并提供材料链接。

**如果你有兴趣成为一名作家**:如果有你擅长的主题，并且你有兴趣写书或投稿，请访问 authors.packtpub.com。



# 复习

请留下评论。一旦你阅读并使用了这本书，为什么不在你购买它的网站上留下评论呢？潜在的读者可以看到并使用您不带偏见的意见来做出购买决定，我们 Packt 可以了解您对我们产品的看法，我们的作者可以看到您对他们的书的反馈。谢谢大家！

更多关于 Packt 的信息，请访问 packtpub.com。