

# 零、前言

SQL Server 从 SQL Server 2016 开始支持机器学习功能。以前被称为 SQL Server 2016 R 服务的 SQL Server 2017 机器学习服务有两种风格，R 和 Python。这本书为数据专业人员、数据库管理员和数据科学家提供了实践参考和学习材料，介绍如何使用 SQL Server with R 安装、开发、部署、维护和管理数据科学和高级分析解决方案。无论您是 SQL Server 新手，还是经验丰富的 SQL Server 专业人员，*Hands-on SQL Server Machine Learning Services with R*包含实用的解释、提示和示例，使您能够最大限度地保持数据分析，以提高效率和安全性。



# 这本书是给谁的

本书面向具有一些或没有 R 经验的数据分析师、数据科学家和数据库管理员，他们渴望在日常工作或使用 SQL Server 的未来项目中轻松交付实用的数据科学解决方案。



# 这本书涵盖的内容

[第 1 章](part0020.html#J2B80-e3f81285367248f4bbc6431bcd4f926d)，*介绍 R 和 SQL Server，*开始我们在 SQL Server 2016 之前的 SQL Server 中的数据科学之旅，并为我们带来今天的 SQL Server R 集成。

[第 2 章](part0025.html#NQU20-e3f81285367248f4bbc6431bcd4f926d)，*Microsoft Machine Learning Server 和 SQL Server 概述，*给出了 Microsoft Machine Learning Server 的简要概述和概述，重点介绍了 SQL Server Machine Learning Services，同时探讨了它是如何工作的以及 R 环境的不同版本。这包括对其背后的架构、不同的计算环境、系统之间的集成如何工作以及如何实现并行性和负载分布的关键讨论。

[第 3 章](part0039.html#1565U0-e3f81285367248f4bbc6431bcd4f926d)，*管理 SQL Server 2017 和 R 的机器学习服务*，涵盖安装和设置，包括如何使用 PowerShell。它包括探索资源调控器的功能、为用户设置角色和安全性以使用 SQL Server Machine Learning Services with R、使用会话和日志、为数据分析或预测建模安装任何缺失或附加的 R 包，以及使用`sp_execute_external_script`外部过程迈出第一步。

[第 4 章](part0057.html#1MBG20-e3f81285367248f4bbc6431bcd4f926d)、*数据探索和数据可视化*，探索数据浏览、分析、管理的 R 语法，以及可视化和预测分析的争论。开发这些技术对于本章和本书的后续步骤至关重要。本章介绍了用于可视化和预测建模的各种有用的 R 包。此外，读者还将了解如何将 R 与 Power BI、SQL Server Reporting Services(SSRS)和移动报告集成。

[第 5 章](part0081.html#2D7TI0-e3f81285367248f4bbc6431bcd4f926d)， *RevoScaleR 包*，讨论了在大型数据集上使用 RevoScaleR 进行可扩展和分布式统计计算的优势。使用 RevoScaleR 提高了 CPU 和 RAM 的利用率，提高了性能。本章向读者介绍 RevoScaleR 在数据准备、描述性统计、统计测试、抽样以及预测建模方面的功能。

[第 6 章](part0096.html#2RHM00-e3f81285367248f4bbc6431bcd4f926d)、*预测建模*，重点帮助第一次踏入预测建模世界的读者。使用 SQL Server 和 SQL Server Machine Learning Services with R，读者将了解如何创建预测、执行数据建模、探索 RevoScaleR 和其他包中可用的高级预测算法，以及如何轻松部署模型和解决方案。最后，调用和运行预测，并将结果暴露给不同的专有工具(如 Power BI、Excel 和 SSRS ),从而完成预测建模。

[第 7 章](part0102.html#318PC0-e3f81285367248f4bbc6431bcd4f926d)，*操作 R 代码*，提供操作 R 代码和 R 预测的提示和技巧。读者将了解到稳定和可靠的流程对于生产中组合 R 代码、持久数据和预测模型的重要性。在本章中，读者将有机会探索采用现有的 R 代码和创建新的 R 代码的方法，然后通过各种现成的客户端工具(如 SQL Server Management Studio (SSMS)和 Visual Studio)将其集成到 SQL Server 中。此外，本章还介绍了读者如何使用 SQL Server 代理作业、存储过程、CLR。NET，和 PowerShell 来产品化 R 代码。

[第 8 章](part0133.html#3UQQQ0-e3f81285367248f4bbc6431bcd4f926d)、*部署、管理和监控包含 R 代码的数据库解决方案*，涵盖了如何在集成 R 代码时管理部署和更改对数据库部署的控制。本章提供了关于如何进行解决方案的集成部署以及如何实现持续集成的指南，包括自动化部署以及如何管理版本控制。在这里，读者将学习有效的方法来监控解决方案，监控代码的有效性，以及解决方案部署后的预测模型。

[第 9 章](part0152.html#4GULG0-e3f81285367248f4bbc6431bcd4f926d)、*面向数据库管理员的 R 机器学习服务*，研究和探索数据库管理员每天、每周和每月执行的任务的监控、性能和故障排除。本章使用简单的示例说明 R 服务对 SQL Server 中涉及的其他角色也很有用，并展示了集成在 SQL Server 中的 R 服务如何通过将基本的监控活动发展成更有用的可操作预测来增强数据库管理员的能力。

[第 10 章](part0158.html#4MLOS0-e3f81285367248f4bbc6431bcd4f926d)、 *R 和 SQL Server 2016/2017 特性扩展*，介绍了 SQL Server 2016 和 2017 的新特性以及 R 服务如何结合使用，例如利用 R 语言的新 JSON 格式，使用对内存 OLTP 技术的新改进来提供几乎实时的分析，将列存储索引和 R 中的新特性相结合，以及如何充分利用它们。它还考虑了如何利用 PolyBase 和 Stretch DB 来超越内部部署，实现混合和云的可能性。最后，查询存储保存了许多来自执行计划的统计数据，R 是执行更深入分析的完美工具。



# 从这本书中获得最大收益

为了使用 SQL Server 机器学习服务并运行本书中的代码示例，需要以下软件:

*   SQL Server 2016 和/或 SQL Server 2017 开发人员版或企业版
*   SQL Server Management Studio (SSMS)
*   R IDE，如 R Studio 或带有 RTVS 扩展的 Visual Studio 2017
*   安装了以下扩展的 Visual Studio 2017 社区版:
    *   Visual Studio 的 r 工具(RTVS)
    *   SQL Server 数据工具(SSDT)
*   VisualStudio.com 在线账户

本书中的章节介绍了软件的安装和配置步骤。



# 下载示例代码文件

你可以从你在[www.packtpub.com](http://www.packtpub.com)的账户下载本书的示例代码文件。如果你在其他地方购买了这本书，你可以访问 www.packtpub.com/support[网站](http://www.packtpub.com/support)并注册，让文件直接通过电子邮件发送给你。

您可以按照以下步骤下载代码文件:

1.  在[www.packtpub.com](http://www.packtpub.com/support)登录或注册。
2.  选择支持选项卡。
3.  点击代码下载和勘误表。
4.  在搜索框中输入图书名称，然后按照屏幕指示进行操作。

下载文件后，请确保使用最新版本的解压缩或解压文件夹:

*   WinRAR/7-Zip for Windows
*   适用于 Mac 的 Zipeg/iZip/UnRarX
*   用于 Linux 的 7-Zip/PeaZip

该书的代码包也托管在 GitHub 上，地址为[https://GitHub . com/packt publishing/SQL-Server-2017-Machine-Learning-Services-with-R](https://github.com/PacktPublishing/SQL-Server-2017-Machine-Learning-Services-with-R)。我们在 https://github.com/PacktPublishing/也有丰富的书籍和视频目录中的其他代码包。看看他们！



# 下载彩色图像

我们还提供了一个 PDF 文件，其中有本书中使用的截图/图表的彩色图像。可以在这里下载:[http://www . packtpub . com/sites/default/files/downloads/SQL server 2017 machinelearningservices withr _ color images . pdf](http://www.packtpub.com/sites/default/files/downloads/SQLServer2017MachineLearningServiceswithR_ColorImages.pdf)。



# 使用的惯例

本书通篇使用了许多文本约定。

`CodeInText`:表示文本中的码字、数据库表名、文件夹名、文件名、文件扩展名、路径名、伪 URL、用户输入和 Twitter 句柄。下面是一个例子:“为了计算交叉表——两个(或更多)变量之间的关系——我们将使用两个函数:`rxCrossTabs`和`rxMargins`

代码块设置如下:

```r
> df <- data.frame(unlist(var_info)) 
> df 
```

任何命令行输入或输出都按如下方式编写:

```r
EXECsp_execute_external_script
          @language = N'R'
          ,@script = N'
                      library(RevoScaleR)
    df_sql <- InputDataSet 
                      var_info <- rxGetInfo(df_sql)
                      OutputDataSet <- data.frame(unlist(var_info))'
                      ,@input_data_1 = N'
    SELECT 
                       BusinessEntityID
                      ,[Name]
                      ,SalesPersonID
                      FROM [Sales].[Store]'
```

**粗体**:表示一个新术语、一个重要的单词或您在屏幕上看到的单词。例如，菜单或对话框中的单词出现在文本中，如下所示。下面是一个例子:“如果设置为 1，您可以随时检查启用的外部脚本的`run_value`。”

警告或重要提示如下所示。

提示和技巧是这样出现的。



# 取得联系

我们随时欢迎读者的反馈。

总体反馈:发送电子邮件`feedback@packtpub.com`，在邮件主题中提及书名。如果你对这本书的任何方面有疑问，请发邮件到`questions@packtpub.com`给我们。

**勘误表**:虽然我们已经尽力确保内容的准确性，但错误还是会发生。如果你在这本书里发现了一个错误，请告诉我们，我们将不胜感激。请访问[www.packtpub.com/submit-errata](http://www.packtpub.com/submit-errata)，选择您的图书，点击勘误表提交表单链接，并输入详细信息。

**盗版**:如果您在互联网上遇到我们作品的任何形式的非法拷贝，如果您能提供我们的地址或网站名称，我们将不胜感激。请通过`copyright@packtpub.com`联系我们，并提供材料链接。

**如果你有兴趣成为一名作家**:如果有一个你擅长的主题，并且你有兴趣写书或投稿，请访问[authors.packtpub.com](http://authors.packtpub.com/)。



# 复习

请留下评论。一旦你阅读并使用了这本书，为什么不在你购买它的网站上留下评论呢？潜在的读者可以看到并使用您不带偏见的意见来做出购买决定，我们 Packt 可以了解您对我们产品的看法，我们的作者可以看到您对他们的书的反馈。谢谢大家！

有关 Packt 的更多信息，请访问 packtpub.com。