

# 零、前言

**监督机器学习**被广泛用于金融、在线广告和分析等领域，因为它允许您训练您的系统进行价格预测、活动调整、客户推荐等，使系统能够自我调整并自行做出决策。这可能带来的好处使得了解机器如何在引擎盖下学习变得至关重要。

这本书将通过许多流行的监督机器学习算法的实现和细微差别来指导你。您将通过快速概述开始这一旅程，并了解有监督的机器学习与无监督的学习有何不同。之后，我们将探索参数模型，如线性和逻辑回归，非参数方法，如决策树，以及各种聚类技术，以促进决策和预测。随着我们的继续，你将使用推荐系统，它被在线公司广泛用于增加用户互动和促进潜在销售。最后，我们将简要介绍一下神经网络和迁移学习。

到本书结束时，你将具备动手技术，获得快速有效地将监督学习算法应用于新问题所需的实用知识。



# 这本书是给谁的

这本书是为有抱负的机器学习开发者准备的，他们想开始有监督的学习。需要具备 Python 编程的中级知识和监督学习的一些基础知识。



# 这本书涵盖的内容

[第 1 章](a1933cd1-a636-44f9-9729-f9e68d9c726e.xhtml)，*迈向监督学习的第一步*，涵盖了监督机器学习的基础知识，让你做好准备开始自己解决问题。这一章包括四个重要部分。首先，我们将设置我们的 Anaconda 环境，并确保我们能够运行这些示例。在接下来的几节中，在我们开始实现最后一节中的算法之前，我们将涉及机器学习背后的更多理论，在最后一节中，我们将设置我们的 Anaconda 环境。

[第 2 章](b51fa9b6-9158-4bcf-9998-18a8f91d3d06.xhtml)，*实现参数模型*，深入探讨参数建模家族中几种流行的监督学习算法。我们将从正式介绍参数模型开始这一部分，然后我们将特别关注两个非常流行的参数模型:线性和逻辑回归。我们将花一些时间理解内部工作原理，然后跳入 Python 并从头开始实际编码。

[第 3 章](028b1786-df10-4e2b-96be-541675edd2cd.xhtml)，*使用非参数模型*，探索非参数模型族。我们将首先讨论偏差-方差的权衡，并解释参数模型和非参数模型在基本层面上的区别。然后我们将进入决策树和聚类方法。最后，我们将讨论非参数模型的一些优点和缺点。

[第 4 章](378fd619-42ca-448d-a0a4-8419697ebcb1.xhtml)，*监督式 ML 中的高级主题*，将时间分为两个主题:推荐系统和神经网络。我们将从协同过滤开始，然后讨论将基于内容的相似性集成到您的协同过滤系统中。最后，我们将进入神经网络和迁移学习。



# 从这本书中获得最大收益

您将需要以下软件来顺利浏览这些章节:

*   Jupyter 笔记本
*   蟒蛇
*   计算机编程语言



# 下载示例代码文件

你可以从你在[www.packt.com](http://www.packt.com)的账户下载本书的示例代码文件。如果你在别处购买了这本书，你可以访问 www.packt.com/support 并注册，让文件直接通过电子邮件发送给你。

您可以按照以下步骤下载代码文件:

1.  在[www.packt.com](http://www.packt.com)登录或注册。
2.  选择支持选项卡。
3.  点击代码下载和勘误表。
4.  在搜索框中输入图书名称，然后按照屏幕指示进行操作。

下载文件后，请确保使用最新版本的解压缩或解压文件夹:

*   WinRAR/7-Zip for Windows
*   适用于 Mac 的 Zipeg/iZip/UnRarX
*   用于 Linux 的 7-Zip/PeaZip

该书的代码包也托管在 GitHub 上，网址是 https://GitHub . com/packt publishing/Supervised-Machine-Learning-with-Python。如果代码有更新，它将在现有的 GitHub 库中更新。

我们在 https://github.com/PacktPublishing/也有丰富的书籍和视频目录中的其他代码包。看看他们！



# 下载彩色图像

我们还提供了一个 PDF 文件，其中有本书中使用的截图/图表的彩色图像。可以在这里下载:[https://www . packtpub . com/sites/default/files/downloads/9781838825669 _ color images . pdf](https://www.packtpub.com/sites/default/files/downloads/9781838825669_ColorImages.pdf)。



# 使用的惯例

本书通篇使用了许多文本约定。

`CodeInText`:表示文本中的码字、数据库表名、文件夹名、文件名、文件扩展名、路径名、伪 URL、用户输入和 Twitter 句柄。下面是一个例子:“将下载的`WebStorm-10*.dmg`磁盘镜像文件挂载为系统中的另一个磁盘。”

代码块设置如下:

```py
from urllib.request import urlretrieve, ProxyHandler, build_opener, install_opener
import requests
import os
pfx = "https://archive.ics.uci.edu/ml/machine-learning databases/spambase/"
data_dir = "data"
```

任何命令行输入或输出都按如下方式编写:

```py
jupyter notebook
```

**粗体**:表示一个新术语、一个重要单词或您在屏幕上看到的单词。例如，菜单或对话框中的单词出现在文本中，如下所示。下面是一个例子:“从管理面板中选择系统信息。”

警告或重要提示如下所示。

提示和技巧是这样出现的。



# 取得联系

我们随时欢迎读者的反馈。

**总体反馈**:如果你对这本书的任何方面有疑问，请在邮件主题中提及书名，并发邮件至`customercare@packtpub.com`联系我们。

**勘误表**:虽然我们已经尽力确保内容的准确性，但错误还是会发生。如果你在这本书里发现了一个错误，请告诉我们，我们将不胜感激。请访问[www.packt.com/submit-errata](http://www.packt.com/submit-errata)，选择您的图书，点击勘误表提交表格链接，并输入详细信息。

**盗版**:如果您在互联网上遇到我们作品的任何形式的非法拷贝，如果您能提供我们的地址或网站名称，我们将不胜感激。请通过`copyright@packt.com`联系我们，并提供材料链接。

**如果你有兴趣成为一名作家**:如果有一个你擅长的主题，并且你有兴趣写一本书或者为一本书投稿，请访问[authors.packtpub.com](http://authors.packtpub.com/)。



# 复习

请留下评论。一旦你阅读并使用了这本书，为什么不在你购买它的网站上留下评论呢？潜在的读者可以看到并使用您不带偏见的意见来做出购买决定，我们 Packt 可以了解您对我们产品的看法，我们的作者可以看到您对他们的书的反馈。谢谢大家！

更多关于 Packt 的信息，请访问[packt.com](http://www.packt.com/)。